/*
 * Copyright (c) 2003 JGoodies Karsten Lentzsch. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  o Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 *  o Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 *  o Neither the name of JGoodies Karsten Lentzsch nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package de.kleopatra.forms.description.jgoodies;

import java.awt.Component;
import java.awt.Container;

import javax.swing.JLabel;

import com.jgoodies.forms.layout.*;

/**
 * c&p from jgoodies.forms: to get the info as return value instead of
 *  dumping to system.out.
 * <p>
 * Provides static methods that help you understand and fix layout problems
 * when using the {@link FormLayout}. Therefore it can dump information
 * about the layout grid, layout groups and cell constraints to the console.
 *
 * @author  Karsten Lentzsch, Jeanette Winzenburg
 * @version $Revision: 1.1.1.1 $
 */
class FormDebugUtils {

	private static String newLine = System.getProperty("line.separator");

	// Console Dump *********************************************************

	/**
	 * Dumps all layout state to the console except grid bounds:
	 * column and row specifications,
	 * column and row groups, cell constraints.
	 */
	public static String getDebugInfo(Container container) {
		return getDebugInfo(container, false);
	}

	/**
	 * Dumps all layout state to the console: column and row specifications,
	 * column and row groups, grid bounds and cell constraints.
	 */
	public static String getDebugInfo(Container container, boolean complete) {
		if (!(container.getLayout() instanceof FormLayout)) {
			return "The container's layout is not a FormLayout." + newLine;
		}
		StringBuffer buf = new StringBuffer();
		FormLayout layout = (FormLayout) container.getLayout();
		buf.append(dumpColumnSpecs(layout));
		buf.append(dumpRowSpecs(layout));
		buf.append(newLine);
		buf.append(dumpColumnGroups(layout));
		buf.append(dumpRowGroups(layout));
		buf.append(newLine);
		buf.append(dumpConstraints(container));
		if (complete) {
			buf.append(dumpGridBounds(container));
		}
		return buf.toString();
	}

	/**
	 * Dumps the layout's column specifications to the console.
	 *
	 * @param layout   the <code>FormLayout</code> to inspect
	 */
	public static String dumpColumnSpecs(FormLayout layout) {
		StringBuffer buf = new StringBuffer();
		appendLine(buf, "COLUMN SPECS:");
		for (int col = 1; col <= layout.getColumnCount(); col++) {
			ColumnSpec colSpec = layout.getColumnSpec(col);
			buf.append(colSpec.toShortString());
			if (col < layout.getColumnCount()) {
				buf.append(", ");
			}
		}
		appendLine(buf, "");
		return buf.toString();
	}

	private static void appendLine(StringBuffer buf, String text) {
		buf.append(text);
		buf.append(newLine);
	}

	/**
	 * Dumps the layout's row specifications to the console.
	 *
	 * @param layout   the <code>FormLayout</code> to inspect
	 */
	public static String dumpRowSpecs(FormLayout layout) {
		StringBuffer buf = new StringBuffer();
		appendLine(buf, "ROW SPECS:   ");
		for (int row = 1; row <= layout.getRowCount(); row++) {
			RowSpec rowSpec = layout.getRowSpec(row);
			buf.append(rowSpec.toShortString());
			if (row < layout.getRowCount()) {
				buf.append(", ");
			}
		}
		buf.append(newLine);
		return buf.toString();
	}

	/**
	 * Dumps the layout's column groups to the console.
	 *
	 * @param layout   the <code>FormLayout</code> to inspect
	 */
	public static String dumpColumnGroups(FormLayout layout) {
		return dumpGroups("COLUMN GROUPS: ", layout.getColumnGroups());
	}

	/**
	 * Dumps the layout's row groups to the console.
	 *
	 * @param layout   the <code>FormLayout</code> to inspect
	 */
	public static String dumpRowGroups(FormLayout layout) {
		return dumpGroups("ROW GROUPS:    ", layout.getRowGroups());
	}

	/**
	 * Dumps the container's grid info to the console if and only
	 * if the container's layout is a <code>FormLayout</code>.
	 *
	 * @param container   the container to inspect
	 * @throws IllegalArgumentException   if the layout is not FormLayout
	 */
	public static String dumpGridBounds(Container container) {
		StringBuffer buf = new StringBuffer();
		appendLine(buf, "GRID BOUNDS");
		buf.append(dumpGridBounds(getLayoutInfo(container)));
		return buf.toString();
	}

	/**
	 * Dumps the grid layout info to the console.
	 *
	 * @param layoutInfo   provides the column and row origins
	 */
	public static String dumpGridBounds(FormLayout.LayoutInfo layoutInfo) {
		StringBuffer buf = new StringBuffer();
		buf.append("COLUMN ORIGINS: ");
		for (int col = 0; col < layoutInfo.columnOrigins.length; col++) {
			buf.append(layoutInfo.columnOrigins[col] + " ");
		}
		buf.append(newLine);
		buf.append("ROW ORIGINS:    ");
		for (int row = 0; row < layoutInfo.rowOrigins.length; row++) {
			buf.append(layoutInfo.rowOrigins[row] + " ");
		}
		buf.append(newLine);
		return buf.toString();
	}

	/**
	 * Dumps the component constraints to the console.
	 *
	 * @param container   the layout container to inspect
	 */
	public static String dumpConstraints(Container container) {
		StringBuffer buf = new StringBuffer();
		appendLine(buf, "COMPONENT CONSTRAINTS");
		if (!(container.getLayout() instanceof FormLayout)) {
			appendLine(buf, "The container's layout is not a FormLayout.");
			return buf.toString();
		}
		FormLayout layout = (FormLayout) container.getLayout();
		int childCount = container.getComponentCount();
		for (int i = 0; i < childCount; i++) {
			Component child = container.getComponent(i);
			CellConstraints cc = layout.getConstraints(child);
			String ccString = (cc == null)
													? "no constraints" : cc.toShortString(layout);
			buf.append(ccString);
			buf.append("; ");
			String childType = child.getClass().getName();
			buf.append(childType);
			if (child instanceof JLabel) {
				JLabel label = (JLabel) child;
				buf.append("      \"" + label.getText() + "\"");
			}
			if (child.getName() != null) {
				buf.append("; name=");
				buf.append(child.getName());
			}
			buf.append(newLine);
		}
		buf.append(newLine);
		return buf.toString();
	}

	// Helper Code **********************************************************

	/**
	 * Dumps the given groups to the console.
	 *
	 * @param title       a string title for the dump
	 * @param allGroups   a two-dimensional array with all groups
	 */
	private static String dumpGroups(String title, int[][] allGroups) {
		StringBuffer buf = new StringBuffer();
		buf.append(title + " {");
		for (int group = 0; group < allGroups.length; group++) {
			int[] groupIndices = allGroups[group];
			buf.append(" {");
			for (int i = 0; i < groupIndices.length; i++) {
				buf.append(groupIndices[i]);
				if (i < groupIndices.length - 1) {
					buf.append(", ");
				}
			}
			buf.append("} ");
			if (group < allGroups.length - 1) {
				buf.append(", ");
			}
		}
		appendLine(buf, "}");
		return buf.toString();
	}

	/**
	 * Computes and answers the layout's grid origins.
	 *
	 * @param container   the layout container to inspect
	 * @return an object that comprises the cell origins and extents
	 * @throws IllegalArgumentException   if the layout is not FormLayout
	 */
	public static FormLayout.LayoutInfo getLayoutInfo(Container container) {
		if (!(container.getLayout() instanceof FormLayout)) {
			throw new IllegalArgumentException(
				"The container must use an instance of FormLayout.");
		}
		FormLayout layout = (FormLayout) container.getLayout();
		return layout.getLayoutInfo(container);
	}
}
