/*
 * Copyright (c) 2003 JGoodies Karsten Lentzsch. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  o Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 *  o Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 *  o Neither the name of JGoodies Karsten Lentzsch nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package de.kleopatra.forms.description.jgoodies;

import java.awt.*;
import java.awt.image.BufferedImage;
import java.util.*;
import java.util.List;

import javax.swing.JPanel;
import javax.swing.SwingUtilities;

import com.jgoodies.forms.layout.FormLayout;

import de.kleopatra.forms.description.LayoutDebugPainter;

/** Painter to show the layoutGrid of a FormLayout on a given Graphics.
 *  <p>
 *  c&p of the painting part from com.jgoodies.forms.debug.FormDebugPanel.
 *
 * @author  Karsten Lentzsch
 * @author  Jeanette Winzenburg
 * @version $Revision: 1.1.1.1 $
 * @see     FormDebugUtils
 */
class FormLayoutDebugPainter implements LayoutDebugPainter {

	/**
	 * The default color used to paint the form's grid.
	 */
	private static final Color DEFAULT_GRID_COLOR = Color.RED;

	private static final Color LEFT_OVERLAY	= Color.RED;
	private static final Color RIGHT_OVERLAY = Color.BLUE;
	/**
	 * Specifies if the grid should be painted in the background.
	 * Is off by default and so the grid is painted in the foreground.
	 */
	private boolean paintInBackground;

	/**
	 * Specifies if the container's diagonals should be painted.
	 */
	private boolean paintDiagonals = true;
	private Container debugee;

	/**
	 * Specifies the color used to paint the grid.
	 */
	private Color gridColor = DEFAULT_GRID_COLOR;
	private String NOT_AVAILABLE = "no debug info available";

	// Instance Creation ****************************************************

	/**
	 * Constructs a <code>FormDebugPanel</code>.
	 * All options are off by default.
	 */
	public FormLayoutDebugPainter() {
		this(null);
	}

	/**
	 * Constructs a <code>FormDebugPanel</code>.
	 * All options are off by default.
	 */
	public FormLayoutDebugPainter(Container debugee) {
		this(debugee, false);
	}

	/**
	 * Constructs a <code>FormDebugPanel</code> for the given
	 * <code>FormLayout</code> and the specified settings.
	 */
	public FormLayoutDebugPainter(Container debugee, boolean paintDiagonals) {

		// super(null);
		setDebugPanel(debugee);
		setPaintDiagonals(paintDiagonals);
		setGridColor(DEFAULT_GRID_COLOR);
	}

//--------------------------implementing LayoutDebugPainter
	public boolean isDebuggable(Container container) {
		return (container != null)
					 && (container.getLayout() instanceof FormLayout);
	}

	public Container getDebugPanel() {
		return debugee;
	}

	public void setDebugPanel(Container cont) {
		this.debugee = cont;
	}

	public void paintDebugInfo(Graphics g) {
		if (isDebuggable(getDebugPanel())) {
			paintGrid(g);
		}
	}

	public void paintDesignHelp(Graphics g) {
	  if (debugee != null) {
	    paintOverlay(g);
	  }
	}
	
	public String getDebugInfo() {
		if (!isDebuggable(getDebugPanel())) {
			return NOT_AVAILABLE;
		}
		return FormDebugUtils.getDebugInfo(getDebugPanel());
	}

	// Accessors ************************************************************

	/**
	 * Enables or disables to paint the panel's diagonals.
	 */
	public void setPaintDiagonals(boolean b) {
		paintDiagonals = b;
	}

	/**
	 * Sets the grid color.
	 *
	 * @param color  the grid color
	 */
	public void setGridColor(Color color) {
		gridColor = color;
	}

	// Painting *************************************************************

	/**
	 * Paints the form's grid lines and diagonals.
	 * pre: isDebuggable(debugee).
	 */
	private void paintGrid(Graphics g) {
		FormLayout layout = (FormLayout) debugee.getLayout();
		FormLayout.LayoutInfo layoutInfo = FormDebugUtils.getLayoutInfo(debugee);
		int left = layoutInfo.getX();
		int top = layoutInfo.getY();
		int width = layoutInfo.getWidth();
		int height = layoutInfo.getHeight();
		g.setColor(gridColor);

		// Paint the column bounds.
		for (int col = 0; col < layoutInfo.columnOrigins.length; col++) {
			g.fillRect(layoutInfo.columnOrigins[col], top, 1, height);
		}

		// Paint the row bounds.
		for (int row = 0; row < layoutInfo.rowOrigins.length; row++) {
			g.fillRect(left, layoutInfo.rowOrigins[row], width, 1);
		}
		if (paintDiagonals) {
			g.drawLine(left, top, left + width, top + height);
			g.drawLine(left, top + height, left + width, top);
		}
	}

	BufferedImage left;
	BufferedImage right;
	
	private void paintOverlay(Graphics g) {
	  updateComponentCaches();
	  Dimension size = debugee.getSize();
	  g.setColor(gridColor);
	  g.drawRect(0, 0, size.width - 1, size.height - 1);
	  g.drawLine(size.width / 2, 0, size.width / 2, size.height );
	  g.setColor(LEFT_OVERLAY);
	  Graphics2D g2;
    BufferedImage left = getOverlayImage(size, getLeftPolygon(size), LEFT_OVERLAY);
	  g.drawImage(left, 0, 0, null);
	  
//	  BufferedImage right = getOverlayImage(size, getRightPolygon(size), RIGHT_OVERLAY);
//	  g.drawImage(right, size.width/2, 0, null);
	  
	}

  private BufferedImage getOverlayImage(Dimension size, Shape polygon, Color overlayColor) {
    BufferedImage left = new BufferedImage(size.width/2, size.height, BufferedImage.TYPE_INT_ARGB_PRE );
	  Graphics2D g2 = left.createGraphics();
	  g2.setComposite(AlphaComposite.getInstance(AlphaComposite.DST_OVER, 0.5f));
	  g2.setColor(overlayColor);
	  g2.fill(polygon);
	  g2.dispose();
    return left;
  }

  private Shape getLeftPolygon(Dimension size) {
    List set = filterHorizontally(topCache, 0, size.width/2);
    List points = new ArrayList();
    Point point = new Point(size.width/2, 0);
    points.add(point);
    for (Iterator iter = set.iterator(); iter.hasNext();) {
      TranslatedLocation component = (TranslatedLocation) iter.next();
      Point p = new Point(point.x, component.leftTop.y);
      points.add(p);
      Point p2 = new Point(component.leftTop.x, p.y);
      points.add(p2);
      point = new Point(p2);
    }
    points.add(new Point(point.x, size.height));
    points.add(new Point(size.width/2, size.height));
    int[] xp = new int[points.size()];
    int[] yp = new int[points.size()];
    for (int i = 0; i < yp.length; i++) {
      xp[i] = ((Point) points.get(i)).x;
      yp[i] = ((Point) points.get(i)).y;
    }
    return new Polygon(xp, yp, points.size());
//    int top = getTopBound(0, size.width/2);
//    int bottom = getBottomBound(0, size.width/2);
//    int left = getLeftBound(0, size.height);
//    return new Rectangle(left, top, size.width/2 - left , size.height - top - bottom);
  }
  
  
	private List filterHorizontally(List cache, int left, int right) {
    List filtered = new ArrayList();
    for (Iterator iter = cache.iterator(); iter.hasNext();) {
      TranslatedLocation component = (TranslatedLocation) iter.next();
      if (!((component.rightBottom.x < left) || (right < component.leftTop.x)) ) {
        filtered.add(component);
      }
     
    }
    return filtered;
  }

	Comparator topComparator;
  List topCache;
	TreeSet leftCache;
	TreeSet rightCache;
	TreeSet bottomCache;
	
  private void updateComponentCaches() {
    if (topCache == null) {
      
      topComparator = new TopComparator();
      topCache = new ArrayList();
      leftCache = new TreeSet(new LeftComparator());
      bottomCache = new TreeSet(new BottomComparator());
    }
    topCache.clear();
    leftCache.clear();
    bottomCache.clear();
    List allChildren = getAllChildren();
    topCache.addAll(allChildren);
    Collections.sort(topCache, topComparator);
    leftCache.addAll(allChildren);
//    bottomCache.addAll(allChildren);
  }
  
  

  private List getAllChildren() {
    List list = new ArrayList();
    addChildren(list, debugee);
    return list;
  }



  private void addChildren(List list, Container container) {
    for (int i = 0; i < container.getComponentCount(); i++) {
      
      Component component = container.getComponent(i);
      if (acceptContainer(component)) {
        addChildren(list, (Container) component);
      }
      if (acceptComponent(component)) {
        list.add(new TranslatedLocation(component));
      }
    }
    
  }



  private boolean acceptComponent(Component component) {
    return ! (component instanceof JPanel) ;
  }

  private boolean acceptContainer(Component component) {
    return component instanceof JPanel;
  }


  private class TranslatedLocation {
    Point leftTop;
    Point rightBottom;
    Component component;
    
    TranslatedLocation(Component comp) {
      setComponent(comp);
    }

    private void setComponent(Component comp) {
      this.component = comp;
     leftTop = SwingUtilities.convertPoint(component.getParent(), component.getLocation(), debugee);
     rightBottom = new Point(leftTop);
     rightBottom.translate(comp.getWidth(), comp.getHeight());
    }
    
    
  }
  private class TopComparator implements Comparator {

    public int compare(Object o1, Object o2) {
      checkComponent(o1);
      checkComponent(o2);
      TranslatedLocation first = (TranslatedLocation) o1;
      TranslatedLocation second = (TranslatedLocation) o2;
      int topFirst = first.leftTop.y;
      int topSecond = second.leftTop.y;
      if (topFirst == topSecond) return 0;
      return topFirst < topSecond ? -1 : 1;
    }

    private void checkComponent(Object o2) {
      if (!(o2 instanceof TranslatedLocation)) throw new IllegalArgumentException("no component");
    }

    
  }

  private class BottomComparator implements Comparator {

    public int compare(Object o1, Object o2) {
      checkComponent(o1);
      checkComponent(o2);
      TranslatedLocation first = (TranslatedLocation) o1;
      TranslatedLocation second = (TranslatedLocation) o2;
      int bottomFirst = first.rightBottom.y;
      int bottomSecond = second.rightBottom.y;
      if (bottomFirst == bottomSecond) return 0;
      return bottomFirst > bottomSecond ? -1 : 1;
    }

    private void checkComponent(Object o2) {
      if (!(o2 instanceof TranslatedLocation)) throw new IllegalArgumentException("no component");
    }

    
  }
  
  private class LeftComparator implements Comparator {

    public int compare(Object o1, Object o2) {
      checkComponent(o1);
      checkComponent(o2);
      TranslatedLocation first = (TranslatedLocation) o1;
      TranslatedLocation second = (TranslatedLocation) o2;
      int topFirst = first.leftTop.x;
      int topSecond = second.leftTop.x;
      if (topFirst == topSecond) return 0;
      return topFirst < topSecond ? -1 : 1;
    }

    private void checkComponent(Object o2) {
      if (!(o2 instanceof TranslatedLocation)) throw new IllegalArgumentException("no component");
    }

    
  }
  

  private int getTopBound(int left, int right) {
    for (Iterator iter = topCache.iterator(); iter.hasNext();) {
      TranslatedLocation component = (TranslatedLocation) iter.next();
      if (!((component.rightBottom.x < left) || (right < component.leftTop.x)) ) {
        return component.leftTop.y;
       // return location.y;
      }
      
    }
    return 0;
  }

  private int getBottomBound(int left, int right) {
    for (Iterator iter = bottomCache.iterator(); iter.hasNext();) {
      TranslatedLocation component = (TranslatedLocation) iter.next();
      if (!((component.rightBottom.x < left) || (right < component.leftTop.x)) ) {
        return component.rightBottom.y;
       // return location.y;
      }
      
    }
    return 0;
  }
  
  private int getLeftBound(int top, int bottom) {
    for (Iterator iter = leftCache.iterator(); iter.hasNext();) {
      TranslatedLocation component = (TranslatedLocation) iter.next();
//      if ((top <= component.leftTop.y) && (component.leftTop.y <= bottom) ) {
    if (!(( component.rightBottom.y < top) || (bottom < component.leftTop.y)) ) {
        return component.leftTop.x;
      }
      
    }
    return 0;
  }
  
  private Shape getRightPolygon(Dimension size) {
    
    return new Rectangle(0, 0, size.width/2, size.height);
  }

}
