package de.kleopatra.forms.description.jgoodies;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.List;
import java.util.StringTokenizer;

import com.jgoodies.forms.layout.CellConstraints;

import de.kleopatra.forms.description.Constraint;
import de.kleopatra.forms.description.LayoutBuilderFactory;

/** Implementation of Constraint for JGoodies Forms.
 *
 * <p>
 * NOTE: this class is public and has public constructors
 * as an implementation side-effect, only. Don't instantiate/access
 * directly from code.
 *  @version $Revision: 1.2 $ $Date: 2004/09/30 10:18:20 $
 *  @author (C) 2003 Jeanette Winzenburg, Berlin
 *
 */
public final class JGoodiesConstraint implements Constraint {

	private static final String DEFAULT_POSITION = "1, 1, ";
	private static final String DEL = ",";
	private CellConstraints constraint;
	private Object componentID;
	private String componentClassName;
	private transient PropertyChangeSupport support;

	/** ussed by XMLDecoder: needed to ensure proper instantiation of wrapped constraint.
	 */
	public JGoodiesConstraint() {
		this((String) null);
	}

	public JGoodiesConstraint(CellConstraints constraint) {
		this.constraint = (CellConstraints) constraint.clone();
	}

	/** encodedString = "column, row, width, height, hAlign, vAlign"
	 *  used by JGoodiesFormFactory
	 */
	public JGoodiesConstraint(String encodedString) {
		setEncodedString(encodedString);
	}

//------------------------------Constraint

	/* (non-Javadoc)
	 * @see de.kleopatra.forms.description.Constraint#addPropertyChangeListener(java.beans.PropertyChangeListener)
	 */
	public void addPropertyChangeListener(PropertyChangeListener l) {
		if (support == null) {
			support = new PropertyChangeSupport(this);
		}
		support.addPropertyChangeListener(l);
	}

	/* (non-Javadoc)
	 * @see de.kleopatra.forms.description.Constraint#removePropertyChangeListener(java.beans.PropertyChangeListener)
	 */
	public void removePropertyChangeListener(PropertyChangeListener l) {
		if (support == null) {
			return;
		}
		support.removePropertyChangeListener(l);
	}

	private void firePropertyChange(String property, Object oldValue,
			Object newValue) {
		if (support == null) {
			return;
		}
		support.firePropertyChange(property, oldValue, newValue);
	}

	public List getHAlignmentTypes() {
		return LayoutBuilderFactory.getLayoutFactory()
			.getConstraintHAlignmentTypes();
	}

	public List getVAlignmentTypes() {
		return LayoutBuilderFactory.getLayoutFactory()
			.getConstraintVAlignmentTypes();
	}

	public Object getHAlignment() {
		return constraint.hAlign;
	}

	public void setHAlignment(Object alignment) {
		CellConstraints.Alignment oldAlign =
			(CellConstraints.Alignment) getHAlignment();
		constraint.hAlign = (CellConstraints.Alignment) alignment;
		firePropertyChange("hAlignment", oldAlign, getHAlignment());
	}

	public Object getVAlignment() {
		return constraint.vAlign;
	}
	;

	public void setVAlignment(Object alignment) {
		CellConstraints.Alignment oldAlign =
			(CellConstraints.Alignment) getVAlignment();
		constraint.vAlign = (CellConstraints.Alignment) alignment;
		firePropertyChange("vAlignment", oldAlign, getVAlignment());
	}

	/** set all properties. The encodedString has the same syntax as
	 * JGoodies CellConstraints.
	 * <p>
	 * Note: this method is meant mainly for handling persistence
	 * and will not fire any event.
	 */
	public void setEncodedString(String encodedString) {
		if ((encodedString != null) && (missingPosition(encodedString))) {
			encodedString = DEFAULT_POSITION + encodedString;
		}
		constraint = createCellConstraints(encodedString);
	}

	public String getEncodedString() {
		String encoded = constraint.toShortString();
		StringTokenizer st = new StringTokenizer(encoded, "(), \"");
		StringBuffer buffer = new StringBuffer(32);
		buffer.append(st.nextToken());
		while (st.hasMoreTokens()) {
			buffer.append(", ");
			buffer.append(st.nextToken());
		}
		return buffer.toString();
	}

	public Object getComponentID() {
		return componentID;
	}

	public void setComponentID(Object componentID) {
		Object old = getComponentID();
		this.componentID = componentID;
		firePropertyChange("componentID", old, getComponentID());
	}

	public String getComponentClassName() {
		return componentClassName;
	}

	public void setComponentClassName(String className) {
		String old = getComponentClassName();
		this.componentClassName = className;
		firePropertyChange("componentClassName", old, getComponentClassName());
	}

	public CellConstraints getCellConstraints() {
		return (CellConstraints) constraint.clone();
	}

	public int getFirstRow() {
		return constraint.gridY - 1;
	}

	public int getFirstColumn() {
		return constraint.gridX - 1;
	}

	public int getLastRow() {
		return getFirstRow() + constraint.gridHeight - 1;
	}

	public int getLastColumn() {
		return getFirstColumn() + constraint.gridWidth - 1;
	}

	public int getHeight() {
		return constraint.gridHeight;
	}

	public int getWidth() {
		return constraint.gridWidth;
	}

	public void extendHeight(int delta) {
		constraint.gridHeight = Math.max(constraint.gridHeight + delta, 1);
	}

	public void extendWidth(int delta) {
		constraint.gridWidth = Math.max(constraint.gridWidth + delta, 1);
	}

	public void extend(int rowDelta, int columnDelta) {
		extendHeight(rowDelta);
		extendWidth(columnDelta);
	}

	public void moveTo(int row, int column) {
		setFirstRow(row);
		setFirstColumn(column);
	}

	public boolean isAnchor(int row, int column) {
		return (getFirstRow() == row) && (getFirstColumn() == column);
	}

	public boolean isInside(int row, int column) {
		return isInsideRowRange(row) && isInsideColumnRange(column);
	}

	public String toString() {
		return getEncodedString();
	}

	public String getDescription() {
		StringBuffer buffer = new StringBuffer();
		buffer.append(constraint.toString());
		buffer.append("; id = ");
		buffer.append(String.valueOf(getComponentID()));
		buffer.append("; class = ");
		buffer.append(getComponentClassName());
		return buffer.toString();
	}

//---------------------------helper
	private void setFirstRow(int row) {
		constraint.gridY = Math.max(row + 1, 1);
	}

	private void setFirstColumn(int column) {
		constraint.gridX = Math.max(column + 1, 1);
	}

	private boolean isInsideRowRange(int row) {
		return (row >= getFirstRow()) && (row <= getLastRow());
	}

	private boolean isInsideColumnRange(int column) {
		return (column >= getFirstColumn()) && (column <= getLastColumn());
	}

	private boolean missingPosition(String encoded) {
		StringTokenizer st = new StringTokenizer(encoded, ",");
		return st.countTokens() < 6;
	}

//---------------------------factory methods
	private CellConstraints createCellConstraints(String encodedString) {
		if (encodedString == null) {
			return new CellConstraints();
		}
		return new CellConstraints(encodedString);
	}
}
