/*
 * Copyright (c) 2003 JGoodies Karsten Lentzsch. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  o Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 *  o Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 *  o Neither the name of JGoodies Karsten Lentzsch nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package de.kleopatra.forms.view;

import java.awt.*;

import javax.swing.JPanel;

import de.kleopatra.forms.description.LayoutDebugPainter;

/** JPanel that can contain a single Component and can
 *  use a LayoutDebugPainter to paint the layoutGrid on top of the comp.
 * <p>
 * refactored and c&p from FormDebugPanel in JGoodies Forms
 *
 *<p>
 *  PENDING: enforce single component
 *
 * @author  Karsten Lentzsch, Jeanette Winzenburg
 * @version $Revision: 1.1.1.1 $
 */
public class FormDebugPanel extends JPanel {

	/**
	 * The default color used to paint the form's grid.
	 */
	private static final Color DEFAULT_GRID_COLOR = Color.red;
	private boolean paintInBackground;
	private Color gridColor;
	private LayoutDebugPainter debugPainter;
	private boolean debugShowing;
  private boolean designHelpShowing;

	// Instance Creation ****************************************************

	/**
	 * Constructs a <code>FormDebugPanel</code>.
	 * All options are off by default.
	 */
	public FormDebugPanel() {
		this(null);
	}

	/**
	 * Constructs a <code>FormDebugPanel</code>.
	 * All options are off by default.
	 */
	public FormDebugPanel(LayoutDebugPainter painter) {
		this(painter, false);
	}

	/**
	 * Constructs a <code>FormDebugPanel</code> for the given
	 * <code>FormLayout</code> and the specified settings.
	 */
	public FormDebugPanel(LayoutDebugPainter debugee,
			boolean paintInBackground) {
		super(new BorderLayout());
		setDebugPainter(debugee);
		setPaintInBackground(paintInBackground);
		setGridColor(DEFAULT_GRID_COLOR);
		setOpaque(false);
	}

	// Accessors ************************************************************
	public void setDebugPainter(LayoutDebugPainter cont) {
		this.debugPainter = cont;
		repaint();
	}

	public LayoutDebugPainter getDebugPainter() {
		return debugPainter;
	}
	public void setDebugShowing(boolean enable) {
		this.debugShowing = enable;
		if (isDesignHelpShowing() && enable) {
		  setDesignHelpShowing(false);
		}
		repaint();
	}

	public boolean isDebugShowing() {
		return debugShowing && (debugPainter != null);
	}

	public void setDesignHelpShowing(boolean enable) {
	  this.designHelpShowing = enable;
	  if (isDebugShowing() && enable) {
	    setDebugShowing(false);
	  }
	  repaint();
	}
	
	public boolean isDesignHelpShowing() {
	  return designHelpShowing;
	}
	
	public void setContent(Container container) {
		removeAll();
		if (debugPainter != null) {
			debugPainter.setDebugPanel(container);
		}
		if (container != null) {
			add(container);
		}
		revalidate();
	}

	/**
	 * Specifies to paint in background or foreground.
	 */
	public void setPaintInBackground(boolean b) {
		paintInBackground = b;
	}

	/**
	 * Sets the grid color.
	 *
	 * @param color  the grid color
	 */
	public void setGridColor(Color color) {
		gridColor = color;
	}

	// Painting *************************************************************

	/**
	 * Paints the component.
	 * @see javax.swing.JComponent#paintComponent(Graphics)
	 */
	public void paintComponent(Graphics g) {
		super.paintComponent(g);
		if (paintInBackground) {
			paintGrid(g);
		}
	}

	/**
	 * Paints the panel. If the panel's layout manager is a
	 * <code>FormLayout</code> it paints the form's grid lines.
	 */
	public void paint(Graphics g) {
		super.paint(g);
		if (!paintInBackground) {
			paintGrid(g);
		}
	}

	/**
	 * Paints the form's grid lines and diagonals.
	 */
	private void paintGrid(Graphics g) {
		if (isDebugShowing()) {
		  debugPainter.paintDebugInfo(g);
		} else if (isDesignHelpShowing()) {
		  debugPainter.paintDesignHelp(g);
		}
		
	}
}
