/*
 * Created on 23.10.2003
 *
 */
package de.kleopatra.forms.view.builder;

import java.awt.Component;
import java.awt.Container;
import java.util.*;

import de.kleopatra.forms.description.*;
import de.kleopatra.support.KleoUtilities;
import de.kleopatra.view.factory.UIFactory;
import de.kleopatra.view.presentation.*;

/** DefaultImplemenation of FormLoader
 *  <p>
 *  uses the shared UIFactory to create/customize containers
 *  and children.
 *
 *
 * @author (C) Jeanette Winzenburg, Berlin
 * @version $Revision: 1.3 $ - $Date: 2004/09/30 10:18:21 $
 */
class DefaultFormLoader implements FormLoader {

  private FormIO formIO;

//----------------------------FormLoader
  public void setFormIO(FormIO formIO) {
    this.formIO = formIO;
  }

  public Container load(String resourceName, Class targetClass) {
    FormDescription formDescription = getFormDescription(resourceName,
                                        targetClass);
    if (formDescription == null) {
      return null;
    }
    Container form = load(formDescription, getUIFactory(), targetClass, true);
    if (form != null) {
      findAndLoadPlaceHolders(form, formDescription, targetClass);

// PENDING: ButtonBar supports needs more analysis... 
//      findAndLoadButtonPanels(form, formDescription.getLayoutDescription());
    }
    return form;
  }

  public Container load(Container form, String resourceName,
      Class targetClass) {
    FormDescription formDescription = getFormDescription(resourceName,
                                        targetClass);
    if (formDescription == null) {
      return null;
    }
    fill(form, formDescription, getUIFactory(), targetClass, true);
    findAndLoadPlaceHolders(form, formDescription, targetClass);
    return form;
  }

  public void rebuild(Container container, Class targetClass) {
    FormDescription formDescription = getFormDescription(container.getName(),
                                        targetClass);
    if (formDescription != null) {
      rebuild(container, formDescription, getUIFactory(), targetClass, false);
    }
  }

  public void rebuildAll(Container container, Class targetClass,
      boolean customize) {
    FormDescription formDescription = getFormDescription(container.getName(),
                                        targetClass);
    if (formDescription == null) {
      return;
    }
    rebuild(container, formDescription, getUIFactory(), targetClass,
            customize);
    List placeHolders =
      findPlaceHoldersNames(formDescription.getLayoutDescription());
    for (Iterator iter = placeHolders.iterator(); iter.hasNext(); ) {
      String element = (String) iter.next();
      Component placeHolder = KleoUtilities.findChildByName(container,
                                element);
      if (placeHolder instanceof Container) {
        FormDescription placeHolderDesc =
          getFormIO().getFormDescription(element, targetClass);
        rebuild((Container) placeHolder, placeHolderDesc, getUIFactory(),
                targetClass, customize);
      }
    }
  }

  public void replaceComponent(Container container, Class targetClass,
      Component replacingComponent) {
    List list = createList();
    list.add(replacingComponent);
    replaceComponents(container, targetClass, list);
  }

  public void replaceComponents(Container container, Class targetClass,
      Collection replacingComponents) {
    FormDescription formDescription = getFormDescription(container.getName(),
                                        targetClass);
    if (formDescription != null) {
      replaceComponents(container, formDescription.getLayoutDescription(),
          replacingComponents);
    }
  }

  public Container loadTemplate(String resourceName, Class targetClass,
      Collection children) {
    Container form = findAndLoad(resourceName, getUIFactory(), targetClass,
                       false);
    if (form != null) {
      removeUnused(form, children);
      replaceComponents(form, targetClass, children);
    }
    return form;
  }

  public void customizeActions(List actions, String resourceName,
      Class targetClass) {
    FormDescription desc = getFormIO().getFormDescription(resourceName,
                             targetClass);
    if (desc != null) {
      Presentation presentation = desc.getPresentation();
      FormFactory.getBundleHandler().localize(desc.getPresentation(),
          targetClass);
      ActionHelper.customizeFromPresentations(actions,
          presentation.getPresentationDescriptions(), targetClass);
    }
  }

  public StyleMap loadStyleMap(String resourceName, Class targetClass) {
    return getFormIO().getStyleMap(resourceName, targetClass);
  }

//-----------------------actual building

  /** create, fill and optionally customize a container.
   *  <p>
   *  use the LayoutDescription and the factory for creation.
   *  if customize use the given Presentation and factory to customize
   *  the container using resources relative to targetClass.
   *
   */
  protected Container load(FormDescription formDescription,
      UIFactory factory, Class targetClass, boolean customize) {
    Container container = create(formDescription.getLayoutDescription(),
                            factory);
    if (customize) {
      customize(container, formDescription.getPresentation(),
                factory.getPresentationManager(), targetClass);

//      factory.getPresentationManager().customize(container,
//          formDescription.getPresentation(), targetClass);
    }
    return container;
  }

  private Container fill(Container container,
      FormDescription formDescription, UIFactory factory, Class targetClass,
      boolean customize) {
    getUIBuilder().fill(container, formDescription.getLayoutDescription(),
        factory);
    if (customize) {
      customize(container, formDescription.getPresentation(),
                factory.getPresentationManager(), targetClass);

//      factory.getPresentationManager().customize(container,
//          formDescription.getPresentation(), targetClass);
    }
    return container;
  }

  /** create default container using the specified UIFactory and fill with
   * children created by the uiFactory and LayoutDescription.
   * <p>
   * PRE: factory != null
   */
  protected Container create(LayoutDescription description,
      UIFactory factory) {
    Container container =
      factory.createDefaultContainer(description.getLayoutID());
    getUIBuilder().fill(container, description, factory);
    return container;
  }

  private void rebuild(Container container, FormDescription formDescription,
      UIFactory factory, Class targetClass, boolean customize) {
    getUIBuilder().refill(container, formDescription.getLayoutDescription());
    if (customize) {
      customize(container, formDescription.getPresentation(),
                factory.getPresentationManager(), targetClass);
    }
    container.invalidate();
    container.validate();
  }

  private void customize(Container container, Presentation presentation,
      PresentationManager presentationManager, Class targetClass) {
    FormFactory.getBundleHandler().localize(presentation, targetClass);
    presentationManager.customize(container, presentation, targetClass);
  }

  /** PRE: all parameters != null.
   *
   * @param container
   * @param formDescription
   * @param replacingComponents
   */
  private void replaceComponents(Container container,
      LayoutDescription layoutDescription, Collection replacingComponents) {
    getUIBuilder().replaceComponents(container, layoutDescription,
        replacingComponents);
    container.invalidate();
    container.validate();
  }

  /** find a resource, create, fill and optionally customize a container with it.
   *  <p>
   *  resources are searched relative to targetClass. If found the container is
   *  created filled and optionally customized. If not found null is returned.
   */
  private Container findAndLoad(String descriptionName, UIFactory factory,
      Class targetClass, boolean customize) {
    FormDescription formDescription = getFormDescription(descriptionName,
                                        targetClass);
    if (formDescription == null) {
      return null;
    }
    return load(formDescription, factory, targetClass, customize);
  }

  /** build a list of forms from a list of descriptionNames.
   *
   */
  protected List findAndLoadAll(List descriptionNames, UIFactory factory,
      Class targetClass, boolean customize) {
    List forms = createList();
    for (Iterator iterator = descriptionNames.iterator();
        iterator.hasNext(); ) {
      String element = (String) iterator.next();
      Container form = findAndLoad(element, factory, targetClass, customize);
      if (form != null) {
        forms.add(form);
      }
    }
    return forms;
  }

  /** PRE: form != null, formDescription != null.
   *
   * @param description
   * @return
   */
  protected void findAndLoadPlaceHolders(Container form,
      FormDescription formDescription, Class targetClass) {
    List placeHolders =
      findPlaceHoldersNames(formDescription.getLayoutDescription());
    List substituents = findAndLoadAll(placeHolders, getUIFactory(),
                          targetClass, true);
    replaceComponents(form, formDescription.getLayoutDescription(),
        substituents);

    //   findAndLoadPlaceHoldersRecursively(substituents, targetClass);
  }

//  protected void findAndLoadButtonPanels(Container form, LayoutDescription layoutDescription) {
//    List buttonPanels = findButtonBarConstraints(layoutDescription);
//    List substituents = findAndLoadAllButtonsPanels(buttonPanels, getUIFactory(), getClass(), true);
//    replaceComponents(form, layoutDescription, substituents);
//  }

  /** build a list of forms from a list of descriptionNames.
  *
  */
// protected List findAndLoadAllButtonsPanels(List descriptionNames, UIFactory factory,
//     Class targetClass, boolean customize) {
//   List forms = createList();
//   String typePostfix;
//   if (LayoutStyle.getCurrent().isLeftToRightButtonOrder()) {
//     typePostfix = "LToR";
//   } else {
//     typePostfix = "RToL";
//   }
//   for (Iterator iterator = descriptionNames.iterator();
//       iterator.hasNext(); ) {
//     Constraint element = (Constraint) iterator.next();
//     String type = getSubType(element.getComponentClassName());
//     if (type != null) {
//       Container form = findAndLoad(type + typePostfix, factory, targetClass, customize);
//       if (form != null) {
//         form.setName(String.valueOf(element.getComponentID()));
//         forms.add(form);
//       }
//     }
//   }
//   return forms;
// }
  private String getSubType(String componentType) {
    int pos = componentType.lastIndexOf('.');
    if (pos < 0) {
      return null;
    }
    return componentType.substring(pos + 1);
  }

  protected void findAndLoadPlaceHoldersRecursively(List substituents,
      Class targetClass) {
    for (Iterator iterator = substituents.iterator(); iterator.hasNext(); ) {
      Container placeHolder = (Container) iterator.next();
      FormDescription placeHolderDescription =
        getFormDescription(placeHolder.getName(), targetClass);
      if (placeHolderDescription != null) {
        findAndLoadPlaceHolders(placeHolder, placeHolderDescription,
            targetClass);
      }
    }
  }

//-----------------------helper for finding components/constraints

  /** returns a List with the ids of placeHolders in the given description.
   * <p>
   *  PENDING: requires knowledge about Contraint/"PlaceHolder" - should be somewhere else.
   *
   */
  protected List findPlaceHoldersNames(LayoutDescription description) {
    return findNamesOfType("PlaceHolder", description);
  }

//  protected List findButtonBarConstraints(LayoutDescription description) {
//      List buttonBarConstraints = createList();
//      for (Iterator iterator = description.getConstraints().iterator();
//          iterator.hasNext(); ) {
//        Constraint element = (Constraint) iterator.next();
//        String name = String.valueOf(element.getComponentClassName());
//        if (name.startsWith("ButtonBar.")) {
//          buttonBarConstraints.add(element);
//        }
//      }
//    return buttonBarConstraints;
//  }
  protected List findNamesOfType(String type, LayoutDescription description) {
    List placeHolders = createList();
    for (Iterator iterator = description.getConstraints().iterator();
        iterator.hasNext(); ) {
      Constraint element = (Constraint) iterator.next();
      if (type.equals(element.getComponentClassName())) {
        placeHolders.add(String.valueOf(element.getComponentID()));
      }
    }
    return placeHolders;
  }

  /** remove all comps from the form that have no duplicate (with same name) in
   *  collection.
   *
   */
  private void removeUnused(Container form, Collection components) {
    for (int i = form.getComponents().length - 1; i >= 0; i--) {
      Component comp = form.getComponent(i);
      if (!hasReplacee(comp, components)) {
        form.remove(i);
      }
    }
  }

  /** checks whether the comp has an duplicate (with same name) in the collection.
   *
   * @param comp the comp to be replaced
   * @param components collection of components
   * @return true if named comp exists in collection, false if not
   */
  private boolean hasReplacee(Component comp, Collection components) {
    for (Iterator iter = components.iterator(); iter.hasNext(); ) {
      Component element = (Component) iter.next();
      if (comp.getName().equals(element.getName())) {
        return true;
      }
    }
    return false;
  }

  private FormDescription getFormDescription(String resourceName,
      Class targetClass) {
    return (resourceName == null)
           ? null : getFormIO().getFormDescription(resourceName, targetClass);
  }

//----------------------factory methods

  /** returns the used formIO.
   *
   * @return the custom formIO if set or the shared default from FormFactory.
   */
  protected FormIO getFormIO() {
    if (formIO != null) {
      return formIO;
    }
    return FormFactory.getFormIO();
  }

  protected UIBuilder getUIBuilder() {
    return FormFactory.getUIBuilder();
  }

  protected Container createDefaultContainer(UIFactory factory) {
    return factory.createDefaultContainer(null);
  }

  protected Container createForm(UIFactory factory) {
    return createDefaultContainer(factory);
  }

  protected UIFactory getUIFactory() {
    return FormFactory.getUIFactory();
  }

  private Collection createSet() {
    return new HashSet();
  }

  protected List createList() {
    return new ArrayList();
  }

  protected Map createMap() {
    return new HashMap();
  }
}
