package de.kleopatra.forms.view.builder;

import java.util.logging.Logger;

import de.kleopatra.forms.description.LayoutBuilderFactory;
import de.kleopatra.forms.description.LayoutDebugPainter;
import de.kleopatra.view.factory.*;
import de.kleopatra.view.presentation.*;

/** central Factory to get everything needed to
 *  build and customize a form from resources.
 *
 *
 *  @version $Revision: 1.3 $ $Date: 2004/09/30 10:18:21 $
 *  @author (C) 2003 Jeanette Winzenburg, Berlin
 */
public class FormFactory {

  private static UIBuilder uiBuilder;
  private static UIFactory uiFactory;
  private static FormWriter formWriter;
  private static FormReader formReader;
  private static FormIO formIO;
  private static FormLoader formLoader;
  private static ExtFormLoader extFormLoader;
  private static final Logger LOGGER = Logger.getLogger("FormFactory");
  private static BundleHandler bundleHandler;

  /** singleton. but the idea is that we will have project
   *  dependent subclasses at a later stage...
   */
  private FormFactory() {}

  /** returns the shared UIBuilder.
   */
  public static UIBuilder getUIBuilder() {
    if (uiBuilder == null) {
      uiBuilder = createUIBuilder();
    }
    return uiBuilder;
  }

  /** returns the shared default UIFactory.
   *  The shared UIFactory initially has custom maps as defined by
   *  MappingManager.
   *  NOTE: changing the factories XXMaps will be effective application-wide.
   */
  public static UIFactory getUIFactory() {
    if (uiFactory == null) {
      uiFactory = createUIFactory();
    }
    return uiFactory;
  }

  /** returns a new UIFactory with custom maps defined by MappingManager.
   * <p>
   * PENDING: actually a SwingUIFactory created, only the XXMaps are
   * dynamically configurable ...
   *
   */
  public static UIFactory createUIFactory() {
    UIFactory factory = new SwingUIFactory();
    Class mappingClass =
      MappingManager.getInstance().getComponentTypeMapClass();
    try {
      ComponentTypeMap map = (ComponentTypeMap) mappingClass.newInstance();
      factory.setComponentTypeMap(map);
      LOGGER.info("setting ComponentTypeMap: " + mappingClass);
    } catch (Exception ex) {
      LOGGER.info("could not create ComponentTypeMap: " + mappingClass
                  + ". Will use default");
    }
    mappingClass = MappingManager.getInstance().getCustomizerMapClass();
    try {
      CustomizerMap map = (CustomizerMap) mappingClass.newInstance();
      factory.getPresentationManager().setCustomizerMap(map);
      LOGGER.info("setting CustomizerMap: " + mappingClass);
    } catch (Exception ex) {
      LOGGER.info("could not create ComponentTypeMap: " + mappingClass
                  + ". Will use default");
    }
    return factory;
  }

  public static BundleHandler getBundleHandler() {
    if (bundleHandler == null) {
      ;
    }
    {
      bundleHandler = createBundleHandler();
    }
    return bundleHandler;
  }

  /** returns the shared default FormReader.
   *
   */
  public static FormReader getFormReader() {
    if (formReader == null) {
      formReader = createFormReader();
    }
    return formReader;
  }

  /** returns the shared FormWriter.
   *
   * @deprecated use FormLoader instead
   */
  public static FormWriter getFormWriter() {
    if (formWriter == null) {
      formWriter = createFormWriter();
    }
    return formWriter;
  }

  /** returns the shared FormLoader.
   *
   */
  public static FormLoader getFormLoader() {
    if (formLoader == null) {
      formLoader = createFormLoader();
    }
    return formLoader;
  }

  /** returns a shared ExtFormLoader.
   *
   * @return
   */
  public static ExtFormLoader getExtFormLoader() {
    if (extFormLoader == null) {
      extFormLoader = createExtFormLoader();
    }
    return extFormLoader;
  }

  public static ExtFormLoader createExtFormLoader() {
    return new PreviewFormLoader();
  }

  /** returns the shared FormIO.
   *  <p>
   *  NOTE: this is used by default by the shared (Ext)FormLoader and the
   *  FormIOHelper - so setting its IOHandlerMap will be effective
   *  application-wide.
   *
   */
  public static FormIO getFormIO() {
    if (formIO == null) {
      formIO = createFormIO();
    }
    return formIO;
  }

  /** create a new FormIO with handlerMap as defined by MappingManager.
   *
   */
  public static FormIO createFormIO() {
    FormIO formIO = new FormIO();
    Class mapClass = MappingManager.getInstance().getIOHandlerMapClass();
    try {
      FormIOHandlerMap map = (FormIOHandlerMap) mapClass.newInstance();
      formIO.setIOHandlerMap(map);
    } catch (Exception ex) {

      // could not instantiate...
    }
    return formIO;
  }

  /** returns a DebugPainter.
   * <p>
   * PENDING: wrong place.
   */
  public static LayoutDebugPainter createLayoutDebugPainter() {
    return LayoutBuilderFactory.getLayoutTranslator()
      .createLayoutDebugPainter();
  }

  public static FormDescription createFormDescription() {
    return new DefaultFormDescription();
  }

  //---------------------------private factory methods
  private static UIBuilder createUIBuilder() {
    return new FormUIBuilder();
  }

  private static FormReader createFormReader() {
    return new FormReader();
  }

  private static FormWriter createFormWriter() {
    return new FormWriter();
  }

  private static FormLoader createFormLoader() {
    return new DefaultFormLoader();
  }

  private static BundleHandler createBundleHandler() {
    return new DefaultBundleHandler();
  }
}
