package de.kleopatra.forms.view.builder;

import java.io.File;
import java.util.List;

import de.kleopatra.forms.description.LayoutDescription;
import de.kleopatra.forms.description.StyleMap;
import de.kleopatra.persistence.IOHandler;
import de.kleopatra.view.presentation.DefaultPresentation;
import de.kleopatra.view.presentation.Presentation;

/** Common functionality for reading/writing form resources from/to files and
 *  reading from classpath-relative resource locations.
 *  <p>
 *  The actual reading/writing is delegated to a bunch of IOHandlers, mapped by
 *  IOHandlerMap. This hides the format of the storage and makes it pluggable. The
 *  default is to use the XMLEncoder/XMLDecoder and is implemented in AbstractIOHandler
 *  and its specialized descendents for Layout/FormDescription and Presentation.
 *
 *  @version $Revision: 1.3 $ $Date: 2004/09/30 10:18:21 $
 *  @author (C) 2003 Jeanette Winzenburg, Berlin
 *
 *
 */
public class FormIO {

  protected FormIOHandlerMap ioHandlerMap;
  public FormIO() {
    initHandlerMap();
  }

//------------------------customize io handling
  public void setIOHandlerMap(FormIOHandlerMap ioHandlerMap) {
    this.ioHandlerMap = ioHandlerMap;
  }

  public FormIOHandlerMap getIOHandlerMap() {
    return ioHandlerMap;
  }

  public IOHandler getIOHandler(Class handledClass) {
    return ioHandlerMap.getIOHandler(handledClass);
  }

//---------------------------simple file writes

  /** pre: filename != null
   */
  public void persistLayoutDescription(LayoutDescription layoutDescription,
      String filename) {
    ioHandlerMap.getIOHandler(LayoutDescription.class)
      .write(layoutDescription, filename);
  }

  public void persistPresentation(Presentation presentation,
      String filename) {
    ioHandlerMap.getIOHandler(Presentation.class).write(presentation,
        filename);
  }

  public void persistFormDescription(FormDescription form, String filename) {
    ioHandlerMap.getIOHandler(FormDescription.class).write(form, filename);
  }

  public void persistStyleMap(StyleMap map, String filename) {
    ioHandlerMap.getIOHandler(StyleMap.class).write(map, filename);
  }

//---------------------------simple file reads

  /** pre: filename != null
   */
  public LayoutDescription readLayoutDescription(String filename) {
    return (LayoutDescription) ioHandlerMap
      .getIOHandler(LayoutDescription.class).read(filename);
  }

  public Presentation readPresentation(String filename) {
    return (Presentation) ioHandlerMap.getIOHandler(Presentation.class)
      .read(filename);
  }

  public FormDescription readFormDescription(String filename) {
    FormDescription desc =
      (FormDescription) ioHandlerMap.getIOHandler(FormDescription.class)
        .read(filename);
    if (desc == null) {
      LayoutDescription layout = readLayoutDescription(filename);
      if (layout != null) {
        desc = createDefaultFormDescription(layout.getLayoutID());
        desc.setLayoutDescription(layout);
        Presentation presentation = readPresentation(filename);
        if (presentation != null) {
          desc.setPresentation(presentation);
        }
      }
    }
    return desc;
  }

  public StyleMap readStyleMap(String filename) {
    return (StyleMap) ioHandlerMap.getIOHandler(StyleMap.class)
      .read(filename);
  }

//---------------------------temporary (??): if the "higher" abstract is not found, create it from "lower" resources
  private FormDescription createDefaultFormDescription(Object formID) {
    return new DefaultFormDescription(formID);
  }

  protected DefaultPresentation createDefaultPresentation(List descriptions,
      String filename) {
    DefaultPresentation presentation = new DefaultPresentation();
    presentation.setPresentationDescriptions(descriptions);
    return presentation;
  }

//---------------------------simple resource access
  public LayoutDescription getLayoutDescription(String resourceName,
      Class targetClass) {
    return (LayoutDescription) ioHandlerMap
      .getIOHandler(LayoutDescription.class)
        .readResource(resourceName, targetClass);
  }

  public Presentation getPresentation(String resourceName,
      Class targetClass) {
    return (Presentation) ioHandlerMap.getIOHandler(Presentation.class)
      .readResource(resourceName, targetClass);
  }

  public FormDescription getFormDescription(String resourceName,
      Class targetClass) {
    FormDescription descr =
      (FormDescription) ioHandlerMap.getIOHandler(FormDescription.class)
        .readResource(resourceName, targetClass);
    if (descr != null) {
      return descr;
    }

    // fallback for
    LayoutDescription layout = getLayoutDescription(resourceName,
                                 targetClass);
    if (layout != null) {
      FormDescription desc =
        createDefaultFormDescription(layout.getLayoutID());
      desc.setLayoutDescription(layout);
      Presentation presentation = getPresentation(resourceName, targetClass);
      if (presentation != null) {
        desc.setPresentation(presentation);
      }
      return desc;
    }
    return null;
  }

  public StyleMap getStyleMap(String resourceName, Class targetClass) {
    return (StyleMap) ioHandlerMap.getIOHandler(StyleMap.class)
      .readResource(resourceName, targetClass);
  }

  public String extractID(String filename) {
    File file = new File(filename);
    filename = file.getName();
    int lastPoint = filename.lastIndexOf(".");
    if (lastPoint > 0) {
      filename = filename.substring(0, lastPoint);
    }
    return filename;
  }

  public String ensureExtension(String filename, String extension) {
    filename = filename.trim();
    int lastPoint = getExtensionPosition(filename);
    if (lastPoint > 0) {

//      return extension.toLowerCase().equals(filename.substring(lastPoint + 1, filename.length()).toLowerCase());
      filename = filename.substring(0, lastPoint);
    }
    return filename + extension;
  }

  public int getExtensionPosition(String filename) {
    return filename.lastIndexOf(".");
  }

//---------------------------init
  private void initHandlerMap() {
    ioHandlerMap = new FormIOHandlerMap();
  }

//---------------------------factory methods
}
