package de.kleopatra.forms.view.builder;

// for experimentals only
//import com.jgoodies.forms.layout.FormLayout;
//import de.kleopatra.support.KleoUtilities;
import java.awt.Component;
import java.awt.Container;
import java.io.File;
import java.util.*;

import de.kleopatra.forms.description.Constraint;
import de.kleopatra.forms.description.LayoutDescription;
import de.kleopatra.view.factory.UIFactory;
import de.kleopatra.view.presentation.ActionHelper;
import de.kleopatra.view.presentation.Presentation;

/** FormWriter responsible for building/filling of containers.
 *  <p>
 *  <ul>
 *  <li> Basic: Does build/rebuild from a given LayoutDescription
 *  and UIFactory.
 *  <li> Extended: can find and load the description from a
 *  resource located by a resoucename (file or package relative resource).
 *  <li> Extended: can optionally customize the built container
 *  if the found resource includes Presentation as well.
 *  <li> Experimental: recursive form building/customizing
 *  <li> Experimental: replace children/fill from template
 *  </ul>
 *  <p>
 *  Does use a FormIO to do the actual reading/writing. Does use the
 *  shared FormIO from FormFactory if none is set explicitely.
 *
 *  <p>
 *  Resources given by name are searched relative to a given class
 *  (can be null) as implemented in ResourceManager. Resources given
 *  by filepaths are interpreted as absolute file locations.
 *
 *
 *  @version $Revision: 1.2 $ $Date: 2004/09/30 10:18:21 $
 *  @author (C) 2003 Jeanette Winzenburg, Berlin
 *
 *  @deprecated replaced by FormLoader/ExtFormLoader
 *    (v073, 2003/10/24)
 */
public class FormWriter {

  private FormIO formIO;

//------------------------------accessing

  /** set a custom formIO to use.
   * <p>
   * if null the writer uses the shared FormIO as returned by
   * FormFactory.
   */
  public void setFormIO(FormIO io) {
    this.formIO = io;
  }

//-----------------------------Basic: use LayoutDescription/UIFactory

  /** create default container using the shared UIFactory and fill with
   * children created by the shared UIFactory and LayoutDescription.
   */
  public Container buildForm(LayoutDescription description) {
    return buildForm(description, getUIFactory());
  }

  /** create default container using the specified UIFactory and fill with
   * children created by the uiFactory and LayoutDescription.
   *
   * @param description - the LayoutDescription to use.
   * @param factory -factory to use.
   * @return created and filled container.
   */
  public Container buildForm(LayoutDescription description,
      UIFactory factory) {
    return buildForm(factory
      .createDefaultContainer(description
        .getLayoutID()), description, factory);
  }

  /**
   * Fill a given Container with children created by the specified UIFactory
   * and LayoutDescription.
   *
   * <p>
   *  NOTE: any children the container might have are removed.
   *
   * @param container the Container which should be filled with the given layout
   * @param description - the LayoutDescription to use.
   * @param factory -factory to use.
   * @return the given container re-filled with newly created children.
   */
  public Container buildForm(Container container,
      LayoutDescription description, UIFactory factory) {
    getUIBuilder().fill(container, description, factory);
    return container;
  }

  /** re-arrange the contents of a container according to the layoutDescription.
   *  <p>
   *  To work as expected all children of the container need to be named
   *  (interpreted as componentID) and match the id of the constraints in the description.
   *  <p>
   *  CHECK: used where? probably replaced by replaceComponents?
   *
   * @param container - container whose contents needs to be re-arranged
   * @param description - the layoutDescription to use.
   */
  public void rebuildForm(Container container,
      LayoutDescription description) {
    getUIBuilder().refill(container, description);
  }

//----------------------------------create/fill form from resource

  /**
   *  Create container from resource as found by ResourceManager and
   *  customize if requested.
   *
   *  uses the shared uifactory.
   *
   * @param descriptionName - id of the FormDescription.
   * @param targetClass - used to find relative resource locations.
   * @param customize - customizes content if true.
   * @return created, filled and (optionally) customized container.
   */
  public Container buildForm(String descriptionName, Class targetClass,
      boolean customize) {
    return findAndBuildForm(descriptionName, getUIFactory(), targetClass,
        customize);
  }

//--------------------------re-arrange a container 
//   (this is implicitely done from resource)

  /** re-arrange the contents of a container from a resource defined by
   *  the containers name from the shared resource location without
   *  customizing the contents.
   *
   * @deprecated unreliably in finding resources - needs targetClass
   * @param container - the container to re-arrange.
   */
  public void rebuildForm(Container container) {
    rebuildForm(container, (Class) null);
  }

  /** re-arrange the contents of a container from a resource defined by
   *  the containers name from a resource location relative to targetclass
   *  without customizing the contents.
   *
   * @param container - the container to re-arrange.
   * @param targetClass
   */
  public void rebuildForm(Container container, Class targetClass) {
    rebuildForm(container, targetClass, false);
  }

  /** re-arrange the contents of a container according to a resource as defined by
   *  the container's name, found relative to some targetclass and optionally customize.
   *  <p>
   *  To work as expected all children of the container need to be named
   *  (interpreted as componentID) and match the id of the constraints in the description.
   *
   * @param container - container whose contents needs to be re-arranged
   * @param targetClass - to find a relative resource location.
   * @param customize - customize the content if true.
   */
  public void rebuildForm(Container container, Class targetClass,
      boolean customize) {
    String descriptionName = container.getName();
    if (descriptionName == null) {
      return;
    }
    FormDescription formDescription =
      getFormIO().getFormDescription(descriptionName, targetClass);
    if (formDescription != null) {
      rebuildForm(container, formDescription, getUIFactory(), targetClass,
                  customize);
    }
  }

//-----------------------------------create/fill forms from files

  /** create, fill and customize a form from a resource located in the filesystem.
   *
   *  <p>
   *  PENDING: where to look for the relative resources in Presentation?
   *   Client responsible for setting userResourceDir?
   *
   * @param filename - absolute path to resource file
   * @param customize - customizes content if true.
   * @return  created, filled and (optionally) custiomized container.
   */
  public Container buildForm(String filename, boolean customize) {
    return readAndBuildForm(filename, getUIFactory(), null, customize);
  }

//---------------------------------customize actions

  /** customize a list of actions from a resource.
   *  <p>
   *  NOTE: this couples implicitely into Swing if used.
   *
   * @param actions
   * @param resourceName
   * @param targetClass
   */
  public void customizeActions(List actions, String resourceName,
      Class targetClass) {
    FormDescription desc = getFormIO().getFormDescription(resourceName,
                             targetClass);
    if (desc != null) {
      Presentation presentation = desc.getPresentation();
      ActionHelper.customizeFromPresentations(actions,
          presentation.getPresentationDescriptions(), targetClass);
    }
  }

//-----------------------------------experimental: build recursively

  /** experimental: create form from given description and factory, auto-replace
   *  placeHolders one level deep.
   *  Customize replacingComponents depending on customizePlaceHolders.
   *    <p>
   *  assumes the placeholder resources to be in ResourceManager.getUserResourceDir().
   *  It's the client's responsibility to set the userResourceDir if needed.
   *  <p>
   *  PENDING: check which location has precedence!
   *
   * @param description
   * @param factory
   * @param customizePlaceHolders
   * @return
   */
  public Container buildCompleteForm(LayoutDescription description,
      UIFactory factory, boolean customizePlaceHolders) {
    Container container = buildForm(description, factory);
    if (container != null) {
      List placeHolders = findPlaceHoldersNames(description);

      // PENDING: should use given factory instead of default
      // PENDING: where to look for relative resources in PresentationDescription?
      // need some way to find relative to description 
      // but: used in Preview during building descriptions only - so it might be 
      // okay to assume that the client has set the ResourceManager.userResourceDir
      // to some reasonable value... 
      List substituents = findAndBuildForms(placeHolders, factory, null,
                            customizePlaceHolders);
      getUIBuilder().replaceComponents(container, description, substituents);
    }
    return container;
  }

  /**
   * experimental: Create container from resource and
   * auto-replace placeholders one level deep.
   * <P>
   * The methods asssumes that the components (constraints)
   * that should be replaced are of generic type "PlaceHolder"
   * and that their name is the id of the resource they should be
   * replaced with.
   *
   * @param descriptionName
   * @param targetClass
   * @param customize
   * @return
   */
  public Container buildCompleteForm(String descriptionName,
      Class targetClass, boolean customize) {
    FormDescription formDescription =
      getFormIO().getFormDescription(descriptionName, targetClass);
    Container form = null;
    if (formDescription != null) {
      form = buildForm(formDescription, getUIFactory(), targetClass,
          customize);
    }
    if (form != null) {
      List placeHolders =
        findPlaceHoldersNames(formDescription.getLayoutDescription());
      List substituents = findAndBuildForms(placeHolders, getUIFactory(),
                            targetClass, customize);
      replaceComponents(form, targetClass, substituents);
    }
    return form;
  }

  /**
   * experimental: Create container from filename and
   * auto-replace placeholders one level deep.
   * <P>
   * The methods asssumes that the components (constraints) that should be replaced are of
   * generic type "PlaceHolder" and that their name is the id of the resource they should be
   * replaced with. The resources are assumed to be in the same directory as the parent.
   *  <p>
   *  PENDING: where to look for the relative resources in Presentation?
   *   Client responsible for setting userResourceDir?
   *
   * @param filename
   * @param customize
   * @return
   */
  public Container buildCompleteForm(String filename, boolean customize) {
    FormDescription formDescription =
      getFormIO().readFormDescription(filename);
    Container form = null;
    if (formDescription != null) {
      form = buildForm(formDescription, getUIFactory(), null, customize);
    }
    if (form != null) {
      List placeHolders =
        findPlaceHoldersNames(formDescription.getLayoutDescription());
      expandPaths(placeHolders, filename);
      List substituents = readAndBuildForms(placeHolders, getUIFactory(),
                            null, customize);
      replaceComponents(form, null, substituents);
    }
    return form;
  }

//--------------------------------experimental: replace/use template

  /** replace a child of container with newComponent using the layoutDescription
   *  of container - as given by the resource with the container's name.
   *  <p>
   *  The replaced component is the child with the same name as the name of
   *  newComponent.
   *
   * @param container
   * @param targetClass
   * @param replacingComponent
   */
  public void replaceComponent(Container container, Class targetClass,
      Component replacingComponent) {
    List list = createList();
    list.add(replacingComponent);
    replaceComponents(container, targetClass, list);
  }

  /** replace a collection of children in container with a collection of new component
   *  using the layoutDescription
   *  of container - as given by the resource with the container's name.
   *  <p>
   *  The replaced components are the children with the same names as the names of components
   *  in the list.
   *
   * @param container
   * @param targetClass to find a formDescription relative to a classpath if needed
   * @param replacingComponents collection of replacing components
   */
  public void replaceComponents(Container container, Class targetClass,
      Collection replacingComponents) {

    // experimental - but now delegates to UIBuilder
    String descriptionName = container.getName();
    if (descriptionName == null) {
      return;
    }
    FormDescription formDescription =
      getFormIO().getFormDescription(descriptionName, targetClass);
    if (formDescription != null) {
      getUIBuilder().replaceComponents(container,
          formDescription.getLayoutDescription(), replacingComponents);
      container.invalidate();
      container.validate();
    }
  }

  /** experimental: use the resource as template, replacing the created widgets with
   *  those in the collection.
   *
   */
  public Container buildFromTemplate(String resourceName, Class targetClass,
      Collection components) {
    Container form = findAndBuildForm(resourceName, getUIFactory(),
                       targetClass, false);
    if (form != null) {
      removeUnused(form, components);
      replaceComponents(form, targetClass, components);
    }
    return form;
  }

  //---------------------------helper
  // used in I15dWriter experimental
  protected Container buildForm(FormDescription formDescription,
      boolean customize) {
    return buildForm(formDescription, getUIFactory(), null, customize);
  }

/*
  private Container buildForm(FormDescription formDescription, Class targetClass,
      boolean customize) {
    return buildForm(formDescription, getUIFactory(), targetClass, customize);
  }
*/

  /** create, fill and optionally customize a container.
   *  <p>
   *  use the LayoutDescription and the factory for creation.
   *  if customize use the given Presentation and factory to customize
   *  the container using resources relative to targetClass.
   *
   * @param formDescription
   * @param factory
   * @param targetClass
   * @param customize
   * @return
   */
  private Container buildForm(FormDescription formDescription,
      UIFactory factory, Class targetClass, boolean customize) {
    Container container = buildForm(formDescription.getLayoutDescription(),
                            factory);
    if (customize) {
      factory.getPresentationManager().customize(container,
          formDescription.getPresentation(), targetClass);
    }
    return container;
  }

  private void rebuildForm(Container container,
      FormDescription formDescription, UIFactory factory, Class targetClass,
      boolean customize) {
    rebuildForm(container, formDescription.getLayoutDescription());
    if (customize) {
      factory.getPresentationManager().customize(container,
          formDescription.getPresentation(), targetClass);
    }
  }

  /** find a resource, create, fill and optionally customize a container with it.
   *  <p>
   *  resource are searched relative to targetClass. If found the container is
   *  created filled and optionally customized. If not found null is returned.
   */
  private Container findAndBuildForm(String descriptionName,
      UIFactory factory, Class targetClass, boolean customize) {
    FormDescription formDescription =
      getFormIO().getFormDescription(descriptionName, targetClass);
    if (formDescription != null) {
      return buildForm(formDescription, factory, targetClass, customize);
    }
    return null;
  }

  private Container readAndBuildForm(String filename, UIFactory factory,
      Class targetClass, boolean customize) {
    FormDescription formDescription =
      getFormIO().readFormDescription(filename);
    if (formDescription != null) {
      return buildForm(formDescription, factory, targetClass, customize);
    }
    return null;
  }

  /** build a list of forms from a list of descriptionNames.
   *
   * @param descriptionNames
   * @param targetClass
   * @param customize
   * @return
   */
  private List findAndBuildForms(List descriptionNames, UIFactory factory,
      Class targetClass, boolean customize) {
    List forms = createList();
    for (Iterator iterator = descriptionNames.iterator();
        iterator.hasNext(); ) {
      String element = (String) iterator.next();
      Container form = findAndBuildForm(element, factory, targetClass,
                         customize);
      if (form != null) {
        forms.add(form);
      }
    }
    return forms;
  }

  /** build a list of forms from a list of filenames.
   *
   * @param descriptionNames list of filenames
   * @param customize the forms are customized if true
   * @return list of containers
   */
  private List readAndBuildForms(List fileNames, UIFactory factory,
      Class targetClass, boolean customize) {
    List forms = createList();
    for (Iterator iterator = fileNames.iterator(); iterator.hasNext(); ) {
      String filename = (String) iterator.next();
      Container form = readAndBuildForm(filename, factory, targetClass,
                         customize);
      if (form != null) {
        forms.add(form);
      }
    }
    return forms;
  }

  /** returns a List with the ids of placeHolders in the given description
   *
   *  PENDING: requires knowledge about Contraint/"PlaceHolder" - should be somewhere else.
   *
   * @param description
   * @return
   */
  private List findPlaceHoldersNames(LayoutDescription description) {
    List placeHolders = createList();
    for (Iterator iterator = description.getConstraints().iterator();
        iterator.hasNext(); ) {
      Constraint element = (Constraint) iterator.next();
      if ("PlaceHolder".equals(element.getComponentClassName())) {
        placeHolders.add(String.valueOf(element.getComponentID()));
      }
    }
    return placeHolders;
  }

  /** expands the ids to the same filepath as filename.
   *
   * @param placeHolders list of resourceIDs
   * @param filename file with the path to expand to
   */
  private void expandPaths(List placeHolders, String filename) {
    File file = new File(filename);
    String name = file.getName();
    String parent = file.getParent();
    for (ListIterator iterator = placeHolders.listIterator();
        iterator.hasNext(); ) {
      String element = (String) iterator.next();
      File created = new File(parent, element);
      iterator.set(created.getPath());
    }
  }

  /** remove all comps from the form that have no duplicate (with same name) in
   *  collection.
   *
   * @param form
   * @param components
   */
  private void removeUnused(Container form, Collection components) {
    for (int i = form.getComponents().length - 1; i >= 0; i--) {
      Component comp = form.getComponent(i);
      if (!hasReplacee(comp, components)) {
        form.remove(i);
      }
    }
  }

  /** checks whether the comp has an duplicate (with same name) in the collection.
   *
   * @param comp the comp to be replaced
   * @param components collection of components
   * @return true if named comp exists in collection, false if not
   */
  private boolean hasReplacee(Component comp, Collection components) {
    for (Iterator iter = components.iterator(); iter.hasNext(); ) {
      Component element = (Component) iter.next();
      if (comp.getName().equals(element.getName())) {
        return true;
      }
    }
    return false;
  }

  //---------------------------factory methods
  protected UIBuilder getUIBuilder() {
    return FormFactory.getUIBuilder();
  }

  protected Container createDefaultContainer(UIFactory factory) {
    return factory.createDefaultContainer(null);
  }

  protected Container createForm(UIFactory factory) {
    return createDefaultContainer(factory);
  }

/*  protected PresentationManager getPresentationManager() {
    return getUIFactory().getPresentationManager();
  }
*/
  protected UIFactory getUIFactory() {
    return FormFactory.getUIFactory();
  }

  /** returns the used formIO.
   *
   * @return the custom formIO if set or the shared default from FormFactory.
   */
  protected FormIO getFormIO() {
    if (formIO != null) {
      return formIO;
    }
    return FormFactory.getFormIO();
  }

  private Collection createSet() {
    return new HashSet();
  }

  protected List createList() {
    return new ArrayList();
  }

  protected Map createMap() {
    return new HashMap();
  }
}
