package de.kleopatra.forms.view.builder;

import java.awt.Frame;
import java.util.Arrays;

import javax.swing.JDialog;

/**
 * A dialog for editing a Form. Uses a Form to build
 * the editor, registers itself as Handler for the form's
 * ok/cancel buttons.
 * These buttons trigger a commit or flush in the dialog's PresentationModel.
 *
 * This is copied from JGoodies Binding tutorial and adjusted to fit into
 * SwingEmpire.
 *
 * NOTE: this will change with a high probability...
 *
 * @author Karsten Lentzsch
 * @author Jeanette Winzenburg
 * @version $Revision: 1.3 $
 *
 */
public class JFormDialog extends JDialog {

  protected Form editor;
  private boolean canceled;

  /**
   * Constructs a Dialog around the Form. <p>
   * PRE: editor != null
   */
  public JFormDialog(Frame parent, Form editor) {
    super(parent, null, true);
    this.editor = editor;
    registerActions();
    canceled = false;
  }

  /** 
   * Set the target to edit in the Form.
   * Has to be reasonably compatible...
   * 
   * @param target
   */
  public void setTarget(Object target) {
    editor.setTarget(target);
  }

//-------------------- action callbacks
  public void ok() {
    editor.commit();
    canceled = false;
    close();
  }

  public void cancel() {
    editor.flush();
    canceled = true;
    close();
  }

  /**
   * Registers the callback methods with the Form.
   * Registration will effect nothing if the actionIDs
   * are unknown to the Form. <p>
   * here: registers "ok", "cancel" - subclasses should override
   * if they need other method hooks.
   *
   */
  protected void registerActions() {
    editor.registerHandler(Arrays.asList(new String[]{ "ok", "cancel" }),
        this);
  }

//------------------------------ Dialog life cycle

  /**
   * Builds the dialog content, marks it as not canceled and makes it visible.
   */
  public void open() {
    build();
    canceled = false;
    setVisible(true);
  }

  /**
   * Closes the dialog: makes it invisible and disposes it,
   * which in turn releases all required OS resources.
   */
  public void close() {
    dispose();
  }

  /**
   * Checks and answers whether the dialog has been canceled.
   * This indicator is set in #doAccept and #doCancel.
   *
   * @return true indicates that the dialog has been canceled
   */
  public boolean hasBeenCanceled() {
    return canceled;
  }

  // Building ***************************************************************

  /**
   * Builds the dialog's content pane, packs it, sets the resizable property,
   * and locates it on the screen. The dialog is then ready to be opened.<p>
   *
   * Internally delegates build and binding to Form.
   * Subclasses should rarely override this method.
   */
  protected void build() {
    setContentPane(editor.getFormComponent());
    pack();
    setResizable(false);
    FormUtils.locateOnScreenCenter(this);
  }
}
