/*
 * Created on 26.11.2003
 *
 */
package de.kleopatra.forms.view.builder;

import java.util.*;

import de.kleopatra.forms.description.jgoodies.JGoodiesStyleMap;
import de.kleopatra.persistence.ResourceKeys;
import de.kleopatra.persistence.ResourceManager;
import de.kleopatra.view.factory.SwingComponentTypeMap;
import de.kleopatra.view.presentation.SwingCustomizerMap;

/**
 * responsible for handling pluggable XXMaps
 * (ComponentTypeMap, CustomizerMap, IOHandlerMap, StyleMap).
 * <p>
 * reads settings of defaults to be used for app from a property value in
 * "resources/forms.properties" (on classpath - see example), if available,
 * and copies to the appropriate system properties it not yet set.
 *
 * <p>
 * Strategy of getXXMap:
 * <ul>
 * <li>
 * if system property XXMap is set - return it
 * <li>
 * if property set in form.properties - set the system prop and return it
 * <li>
 * if neither is set return hardcoded defaultXX
 * </ul>
 * <p>
 * PENDING: support dynamic switching at runtime?
 * Not available at - needs major refactoring to include
 * notifications about map changes.
 *
 * @author (C) Jeanette Winzenburg, Berlin
 * @version $Revision: 1.2 $ - $Date: 2004/09/30 10:18:21 $
 */
public class MappingManager {

  protected final String defaultComponentTypeMap =
    "de.kleopatra.view.factory.SwingComponentTypeMap";
  protected final String defaultCustomizerMap =
    "de.kleopatra.view.presentation.SwingCustomizerMap";
  protected final String defaultIOHandlerMap =
    "de.kleopatra.forms.view.builder.FormIOHandlerMap";
  protected final Class defaultComponentTypeMapClass =
    SwingComponentTypeMap.class;
  protected final Class defaultCustomizerMapClass = SwingCustomizerMap.class;
  protected final Class defaultIOHandlerMapClass = FormIOHandlerMap.class;
  protected final Class defaultStyleMapClass = JGoodiesStyleMap.class;
  protected String defaultFormsPath = "resources/forms";
  protected ResourceBundle bundle;

//  private List availableCustomizerMaps;
//  private List availableComponentMaps;
//  private List availableIOHanderMaps;
  private static MappingManager mappingManager;
  protected MappingManager() {}

  public static MappingManager getInstance() {
    if (mappingManager == null) {
      mappingManager = new MappingManager();
      mappingManager.init();
    }
    return mappingManager;
  }

  public Class getCustomizerMapClass() {
    return getMapClass(defaultCustomizerMapClass,
        ResourceKeys.CUSTOMIZER_MAP_KEY,
        ResourceKeys.DEFAULT_CUSTOMIZER_MAP_KEY);
  }

  public Class getComponentTypeMapClass() {
    return getMapClass(defaultComponentTypeMapClass,
        ResourceKeys.COMPONENT_MAP_KEY,
        ResourceKeys.DEFAULT_COMPONENT_MAP_KEY);
  }

  public Class getIOHandlerMapClass() {
    return getMapClass(defaultIOHandlerMapClass,
        ResourceKeys.IOHANDLER_MAP_KEY,
        ResourceKeys.DEFAULT_IOHANDLER_MAP_KEY);
  }

  public Class getStyleMapClass() {
    return getMapClass(defaultStyleMapClass, ResourceKeys.STYLE_MAP_KEY,
        ResourceKeys.STYLE_MAP_KEY);
  }

  private Class getMapClass(Class defaultMap, String resourceKey,
      String oldResourceKey) {
    String mapName = System.getProperty(resourceKey);
    if (mapName == null) {
      mapName = System.getProperty(oldResourceKey);
    }
    if ((mapName != null) &&!defaultMap.getName().equals(mapName)) {
      try {
        Class mapClass = ResourceManager.getInstance().getClass(mapName);
        return mapClass;
      } catch (ClassNotFoundException ex) {

        // fallback to default on error
      }
    }
    return defaultMap;
  }

//  public String getCustomizerMapClassName() {
//    String map = System.getProperty(ResourceKeys.DEFAULT_CUSTOMIZER_MAP_KEY);
//    if (map == null) {
//      map = defaultCustomizerMap;
//    }
//    return map;
//  }
//  public void setCustomizerMapClassName(String className) {
//    setSystemProperty(ResourceKeys.CUSTOMIZER_MAP_KEY, className);
//  }
//  public String getComponentTypeMapClassName() {
//    String map = System.getProperty(ResourceKeys.DEFAULT_COMPONENT_MAP_KEY);
//    if (map == null) {
//      map = defaultComponentTypeMap;
//    }
//    return map;
//  }
//  public void setComponentTypeMapClassName(String className) {
//    setSystemProperty(ResourceKeys.COMPONENT_MAP_KEY, className);
//  }
//  public String getIOHandlerMapClassName() {
//    String map = System.getProperty(ResourceKeys.DEFAULT_IOHANDLER_MAP_KEY);
//    if (map == null) {
//      map = defaultIOHandlerMap;
//    }
//    return map;
//  }
//  public void setIOHandlerMapClassName(String className) {
//    setSystemProperty(ResourceKeys.IOHANDLER_MAP_KEY, className);
//  }
//  public List getAvailableCustomizerMapClassNames() {
//    if (availableCustomizerMaps == null) {
//      availableCustomizerMaps = createAndInitAvailableCustomizers();
//    }
//    return availableCustomizerMaps;
//  }
//
//  public List getAvailableComponentTypeMapClassNames() {
//    if (availableComponentMaps == null) {
//      availableComponentMaps = createAndInitAvailableComponents();
//    }
//    return availableComponentMaps;
//  }
//
//  public List getAvailableIOHandlerMapClassNames() {
//    if (availableIOHanderMaps == null) {
//      availableIOHanderMaps = createAndInitAvailableIOHandlerMaps();
//    }
//    return availableIOHanderMaps;
//  }
//-----------------------init
  protected void init() {
    try {
      bundle = ResourceBundle.getBundle(defaultFormsPath,
          Locale.getDefault(),
          Thread.currentThread().getContextClassLoader());

      // update the old properties
      updateSystemProperty(ResourceKeys.DEFAULT_COMPONENT_MAP_KEY);
      updateSystemProperty(ResourceKeys.DEFAULT_CUSTOMIZER_MAP_KEY);
      updateSystemProperty(ResourceKeys.DEFAULT_IOHANDLER_MAP_KEY);

      // update the new properties
      updateSystemProperty(ResourceKeys.COMPONENT_MAP_KEY);
      updateSystemProperty(ResourceKeys.CUSTOMIZER_MAP_KEY);
      updateSystemProperty(ResourceKeys.IOHANDLER_MAP_KEY);
      updateSystemProperty(ResourceKeys.STYLE_MAP_KEY);
    } catch (MissingResourceException ex) {

      // System.out.println("Info: could not find resource: " + defaultFormsPath);
    }
  }

  /** pre: bundle loaded. */
  private void updateSystemProperty(String key) {
    if (System.getProperty(key) != null) {
      return;
    }
    String prop = System.getProperty(key);
    if (prop == null) {
      try {
        prop = bundle.getString(key);
        System.setProperty(key, prop);
      } catch (MissingResourceException ex) {}
    }
  }

  private void setSystemProperty(String key, String className) {
    if (className == null) {
      System.getProperties().remove(key);
    } else {
      System.setProperty(key, className);
    }
  }

//-----------------------------
//  protected List createAndInitAvailableCustomizers() {
//    return addMaps(ResourceKeys.AVAILABLE_CUSTOMIZER_MAPS_KEY, defaultCustomizerMap);
//    
//  } 
//  private List createAndInitAvailableIOHandlerMaps() {
//    return addMaps(ResourceKeys.AVAILABLE_IOHANDLER_MAP_KEY, defaultIOHandlerMap);
//  }
//
//  private List createAndInitAvailableComponents() {
//    return addMaps(ResourceKeys.AVAILABLE_COMPONENT_MAPS_KEY, defaultComponentTypeMap);
//  }
//
//  private List addMaps(String key, String defaultMap) {
//    List list = createList();
//    list.add(defaultMap);
//    List available = getAvailableMaps(key);
//    // PENDING: need to remove duplicates...
//    list.addAll(available);
//    return list;
//  }
//  
//  private List getAvailableMaps(String key) {
//    if (bundle == null) return Collections.EMPTY_LIST;
//    List list = createList();
//    try {
//      String maps = bundle.getString(key);
//      StringTokenizer st = new StringTokenizer(maps, ", ");
//      while (st.hasMoreTokens()) {
//        list.add(st.nextToken());
//      }
//      
//    } catch (MissingResourceException ex) {
//      // could not find key 
//    }
//    return list;
//  }
//
//  protected List createList() {
//    return new ArrayList();
//  }
}
