package de.kleopatra.persistence;

import java.io.File;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;

import javax.swing.Icon;
import javax.swing.ImageIcon;

/** handles a) resources with fallback in look-up and b) dynamic classloading.
 * 
 * <p> 
 * getResource/getResourceAsStream/getIcon all use the same strategy 
 * to find a resource:
 * <ul>  
 * <li> <b>if targetClass == null</b>:
 *  then only lookup relative to userResourceDir
 * <li> <b>if targetClass != null</b>:
 *  then first lookup in the class-relative locations; 
 *  if the class-relative lookup is unsuccessful 
 * 	then lookup relative to userResourceDir 
 * </ul> 
 * 
 *  
 *<p>
 *  the fallback sequence is:
 *  <p>
 * 	<ol> 
 * 	<li> look for static resources (those that are created at compiletime) 
 * 		somewhere	relative to the current classloader 
			(available for targetClass != null only)
 * 		<ol>
 * 		<li> look in classpath of targetClass:
 * 			[packagename]/resourcePrefix/[resourceName]
 * 		<li> look in classpath analogon below shared resource location:  
 * 			/resourcePrefix/[packageName]/[resourceName]
 * 		<li> look in shared resource location: 
 * 			/resourcePrefix/[resourceName]
 * 		</ol>
 * 	
 * 	<li> look for dynamic resources (those created/manipulated at runtime)
 * 		in the file system
 * 		<ol>
 * 	  <li> will be removed: - no need to assume some hardcoded dir
 * 				relative to an explicitly set location 
 * 				look in default resourceDir:  
 * 			 [userResourceDir]/resourcePrefix/[resourceName]
 * 		<li> look in userResourceDir directly: 
 * 			 [userResourceDir]/[resourceName]
 * 		</ol> 
 * </ol>
 * 
 * NOTE: exceptions are caught silently - if one occured the returned 
 * URL/InputStream/Icon is null. That's the same behaviour as in 
 * 	class.getResource()/getResourceAsStream.
 * <p>  
 * PENDING: lookup strategy should be encapsulated in pluggable 
 * 	strategy
 * <p>
 * NOTE: does use context classloader for dynamic class loading. 
 * Framework client code should
 * call ResourceManager.getClass() instead Class.forName().
 * 
 * <p>
 * 	THINK: reversing the outer 1 and 2 would enable 
 * 		dynamic replace of resources at runtime?
 * 
 * @author (C) Jeanette Winzenburg, Berlin
 *  @version $Revision: 1.2 $ - $Date: 2004/09/29 14:40:19 $
 */
public class ResourceManager {

	private static final String DEFAULT_RESOURCE_DIR = "resources/";
	private static ResourceManager manager;
	/** the path element of the shared resource path. */ 
	protected String resourcePrefix = DEFAULT_RESOURCE_DIR;
	private String userResourceDir;
	private ResourceBundle bundle;

	private ResourceManager() {}

	/** returns instance of ResourceManager.
	 */
	public static ResourceManager getInstance() {
		if (manager == null) {
			manager = createResourceManager();
		}
		return manager;
	}

	/** returns an URL with fallback-handling.
	 * 
	 */
	public URL getResource(String resourceName, Class targetClass) {
		URL url = null;
		if (targetClass != null) {
			url = getResourceFromClass(resourceName, targetClass);
		}
		if (url == null) {
			url = getResourceFromFileSystem(resourceName);
			
		}
		return url;
	}

  public ResourceBundle getResourceBundle(String resourceName, Class targetClass) {
		ResourceBundle url = null;
		if (targetClass != null) {
			url = getBundleFromClass(resourceName, targetClass);
		}
		if (url == null) {
			//url = getBundleFromFileSystem(resourceName);
			
		}
		return url;
  }
	
	/** returns an InputStream with fallback-handling.
	 * 
	 */
	public InputStream getResourceAsStream(String resourceName,
			Class targetClass) {
		InputStream in = null;
		if (targetClass != null) {
			in = getResourceAsStreamFromClass(resourceName, targetClass);
		}
		if (in == null) {
			in = getFileInputStream(resourceName);
		}
		return in;
	}

	/** returns an icon with fallback-handling.
	 * 
	 */
	public Icon getIcon(String resourceName, Class targetClass) {
		URL url =getResource(resourceName, targetClass);
		if (url != null) {
			return new ImageIcon(url);
		}
		return null;
	}

	public ImageIcon getImageIcon(String resourceName, Class targetClass) {
		Icon icon = getIcon(resourceName, targetClass);
		if (icon instanceof ImageIcon) {
			return (ImageIcon) icon;
		}
		return null;
	}
	/** set user defined resource directory.
	 *
	 */
	public void setUserResourceDir(String dir) {
		userResourceDir = dir;
	}

	public String getUserResourceDir() {
		return userResourceDir;
	}

	/** sets the classpath relative path for the resources.
	 * <p>
	 * a null value resets to default ( == "resources/"),
	 * a not-empty String is checked and changed to have a slash
	 * at the end, an empty String is used as is.
	 * 
	 * @param prefix
	 */ 
	public void setResourcePrefix(String prefix) {
		if (prefix == null) {
			prefix = DEFAULT_RESOURCE_DIR;
		} else if (!"".equals(prefix) && !prefix.endsWith("/")) {
			prefix += "/";
		}
		resourcePrefix = prefix;
	}
	
	public String getResourcePrefix() {
		return resourcePrefix;
	}
	
	public String getProperty(String key) {
		return System.getProperty(key);
	}

	public void setUserBundlePath(String path, ClassLoader loader) {
		bundle = ResourceBundle.getBundle(path, Locale.getDefault(), loader);
	
	}

	public String getBundleProperty(String key) {
		if (bundle == null) return null;
	
		try {
			return bundle.getString(key);
			
		} catch (MissingResourceException e) {
			return null;
		}

	}

	public Class getClass(String className) throws ClassNotFoundException {
	//	return Class.forName(className, true, getCallerClass(0).getClassLoader());
		return Class.forName(className, true, Thread.currentThread().getContextClassLoader());
	}

	
//------------------------------helper

	/** returns URL relative to targetClass with fallback sequence 1
	 *  PRE: targetClass != null
	 */
	private URL getResourceFromClass(String resourceName, Class targetClass) {
		URL url = targetClass.getResource(getResourcePathBelowClass(resourceName,
								targetClass));
		if (url == null) {
			url = targetClass.getResource(getResourcePathBelowSharedAndClass(resourceName,
					targetClass));
		}
		if (url == null) {
			url = targetClass.getResource(getSharedResourcePath(resourceName));
		}
		return url;
	}

	/** returns ResourceBundle relative to targetClass with fallback sequence 1
	 *  PRE: targetClass != null
	 */
	private ResourceBundle getBundleFromClass(String resourceName, Class targetClass) {
		ResourceBundle bundle = getBundle(getBundlePathBelowClass(resourceName, targetClass), targetClass);
		if (bundle == null) {
			bundle = getBundle(getBundlePathBelowSharedAndClass(resourceName,
					targetClass), targetClass);
		}
		if (bundle == null) {
			bundle = getBundle(getSharedBundlePath(resourceName), targetClass);
		}
		return bundle;
	}
	private ResourceBundle getBundle(String path, Class targetClass) {
	  try {
	    if (targetClass != null) {
	      return ResourceBundle.getBundle(path, Locale.getDefault(), targetClass.getClassLoader());
	    } else {
	      return ResourceBundle.getBundle(path, Locale.getDefault());
	    }
	  } catch (MissingResourceException ex) {
	    
	  }
    return null;
  }

  /** returns InputStream relative to targetClass with fallback sequence 1
	 *  PRE: targetClass != null
	 */
	private InputStream getResourceAsStreamFromClass(String resourceName,
			Class targetClass) {
		InputStream in =
			targetClass.getResourceAsStream(getResourcePathBelowClass(resourceName,
				targetClass));
		if (in == null) {
			in = targetClass
				.getResourceAsStream(getResourcePathBelowSharedAndClass(resourceName,
					targetClass));
		}
		if (in == null) {
			in = targetClass
				.getResourceAsStream(getSharedResourcePath(resourceName));
		}
		return in;
	}
	
	/** returns URL relative to userResourceDir with fallback sequence 2.
	 * 
	 */
	private URL getResourceFromFileSystem(String resourceName) {
		URL url = getFileURLFromUserDir(resourceName, true);
		if (url == null) {
			url = getFileURLFromUserDir(resourceName, false);
		}
		return url;
	}

	/** returns InputStream relative to userResourceDir with fallback sequence 2.
	 *  //will be Changed to:
	 *  no fallback - if userDir is set, use it without modification!
	 */
	private InputStream getFileInputStream(String resourceName) {
	//	return getInputStream(getFileURLFromUserDir(resourceName, false));

		InputStream in = getInputStream(getFileURLFromUserDir(resourceName, true));
		if (in == null) {
			in = getInputStream(getFileURLFromUserDir(resourceName, false));
		}
		return in;

	}

	/** returns an url relativ to userResourceDir if set or null if not set.
	 *  the filepath is build as: 
	 * 	if prefixShared: <userResourceDir>/<resourcePrefix>/resourceName
	 * 	if !prefixShared: <userResourceDir>/resourceName.
	 * 
	 * @param resourceName
	 * @param prefixShared
	 * @return the url if userResourceDir is set and the file exists or 
	 * 	null otherwise (and if malformedexception is detected)
	 */
	private URL getFileURLFromUserDir(String resourceName, boolean prefixShared) {
		if (getUserResourceDir() == null) {
			return null;
		}
		try {
			if (prefixShared) {
				resourceName = getSharedResourcePath(resourceName);
			}
			File file = new File(getUserResourceDir(), resourceName);
			if (file.exists()) {
				return file.toURL();
			}
		} catch (MalformedURLException ex) { // do nothing - can't help it
			//ex.printStackTrace();
		}
		return null;
	}

	/** opens and returns the stream of the url, or null if error.
	 * 
	 * @param url
	 * @return
	 */
	private InputStream getInputStream(URL url) {
		try {
			return url.openStream();
		} catch (Exception e) {
			return null;
		}
	}


//---------------------------helper methods to build path names

	/** returns a class specific resourcename below the shared resource.
	 *  <resourceDir>/<packagename>
	 * 	PRE: targetClass != null
	 */
	private String getResourcePathBelowSharedAndClass(String filename,
			Class targetClass) {
		String	packageName = targetClass.getPackage().getName();
		packageName = packageName.replace('.', '/');
		packageName = "/" + resourcePrefix + packageName + "/" + filename;
		return packageName;
	}

	private String getResourcePathBelowClass(String resourceName,
			Class targetClass) {
		return resourcePrefix + resourceName;
	}

	private String getSharedResourcePath(String resourceName) {
		return "/" + resourcePrefix + resourceName;
	}

	/** returns a class specific resourcename below the shared resource.
	 *  <resourceDir>/<packagename>
	 * 	PRE: targetClass != null
	 */
	private String getBundlePathBelowSharedAndClass(String filename,
			Class targetClass) {
		String	packageName = targetClass.getPackage().getName();
		packageName = resourcePrefix + packageName + "/" + filename;
		return packageName;
	}

	private String getBundlePathBelowClass(String resourceName,
			Class targetClass) {
		return targetClass.getPackage().getName() + "/" + resourcePrefix + resourceName;
	}

	private String getSharedBundlePath(String resourceName) {
	  if ("/".equals(resourcePrefix)) {
	    return resourceName;
	  }
		return resourcePrefix + resourceName;
	}

//------------------not used currently: experiments to prepare safer classloading

	private Class getCallerClass(int offset) {
		// 0 = resolver.getClassContext
		// 1 = resourceManager.getCallerClass
		// 2 = resourceManager.getClass
		// 3 = original caller
		return getCallerResolver().getClassContext()[3 + offset];
	}
	
	private CallerResolver callerResolver;
	
	private CallerResolver getCallerResolver() {
		if (callerResolver == null) {
			callerResolver = new CallerResolver();
		}
		return callerResolver;
	}

	private static class CallerResolver extends SecurityManager {
		boolean done;
		protected Class[] getClassContext() {
/*			if (!done) {
				done = true;
				try {
						throw new RuntimeException("in getClassContext");
				} catch (Exception ex) {
					ex.printStackTrace();
				}
			}
*/
			return super.getClassContext();
		}
	}

//---------------------------factory methods
	private static ResourceManager createResourceManager() {
		return new ResourceManager();
	}

}
