package de.kleopatra.view.binding;

import java.util.*;

import javax.swing.*;

import com.jgoodies.binding.beans.BeanAdapter;
import com.jgoodies.binding.value.*;

import de.kleopatra.support.KleoUtilities;

/**
 * Helper to manage data binding.
 * <p>
 * Finds and binds input widgets of source container to properties of 
 * a bean-like target.
 * Name of widgets is used as identifier of the widget.<p>
 *
 * Has state: exactly one container and exactly one target! Internally
 * delegates into both directions: 
 * <ul>
 * <li> to com.jgoodies.binding.bean.BeanAdapter to get ValueModels
 * bound to bean properties. 
 * <li> to a BindingHandler as returned from BindingMap to bind this ValueModel
 * to a component.
 * </ul>
 * 
 * 
 * Conventions for bindXXInput methods:<ul>
 * <li> methods taking one String parameter assume that the property to bind
 * and the componentName are the same
 * <li> in methods with two String parameters the first is the componentName and
 * the second is the propertyName
 * <li> usually, any models passed in are adapted somehow
 * <li> if a widget supports the editable property, the method without explicitly
 * stating the editability will use the default (f.i. true for TextFields, false
 * for ComboBoxes
 * <li> if the component cannot be found (or has the wrong type) a BindingException
 * is thrown. <strong> NOTE: </strong> this is a change of behaviour compared
 * to earlier versions! 
 * </ul>
 *
 * PENDING:
 * As an alternativ to the explicit bindXXInput, binding can be
 * done by providing a collection of BindingMetaData. Not yet complete...
 * <p>
 * 
 * PENDING: 
 * better support external triggers? <p>
 * 
 * PENDING:
 * think about using PresentationModel? <p>
 * 
 * @author (C) Jeanette Winzenburg, Berlin
 * @version $Revision: 1.8 $ - $Date: 2004/10/04 11:56:04 $
 *
 *
 */
public class BeanBinding {

  /** used for internally created buffered adapters. */
  protected Trigger internalTrigger;
  /** externally accessible trigger. */
  protected Trigger externalTrigger;
  /** adapter vending valueModels bound to target properties. */
  protected BeanAdapter beanAdapter;

  /** flag to control listening to the adapted bean */
  protected boolean observeChanges;

  /** the common parent for all widgets to bind. */
  protected JComponent input;

  /** a map of components to enable/disable dependent on
   *  a button's selection state.
   *  THINK: put somewhere else?
   */
  private Map multiplexers;
  private boolean triggering;

// does not make sense to have binding without container -
// all the binding methods will do nothing!
//  public BeanBinding(Object target) {
//    if (target != null) {
//      setTarget(target);
//    }
//  }

  /**
   * Prepares binding between children of input container
   * with listening option. The initial target bean is null.
   *
   * @param input - the parent container of the widgets to bind.
   *  Must not be null.
   * @param observeChanges - flag to indicate if the adapter should
   *  listen to changes of the target bean.
   */
  public BeanBinding(JComponent input, boolean observeChanges) {
    this(input, null, observeChanges);
  }

  /**
   * Prepares binding between children of input container and
   *  target bean without listening to changes of target bean.
   *
   * @param input - the parent container of the widgets to bind.
   *  Must not be null.
   * @param target - the target bean to bind
   */
  public BeanBinding(JComponent input, Object target) {
    this(input, target, false);
  }

  /**
   * Prepares binding between children of input container and
   *  target bean with listening option.
   *
   * @param input - the parent container of the widgets to bind.
   *  Must not be null.
   * @param target - the target bean to bind, allowed to be null.
   * @param observeChanges - flag to indicate if the adapter should
   *  listen to changes of the target bean.
   */
  public BeanBinding(JComponent input, Object target,
      boolean observeChanges) {
    if (input == null) {
      throw new IllegalArgumentException(
        "The input container must not be null");
    }
    this.observeChanges = observeChanges;
    this.input = input;
    if (target != null) {
      setTarget(target);
    }
  }

//------------------ configure (before binding anything)
  
  /** 
   * flag to enable buffering of ValueModels. 
   * if true changes in components will be written
   * to bean properties on commit() only and can be
   * discarded by calling flush(). By default, buffering
   * is disabled. <p>
   * 
   * NOTE: this must be done before calling any
   * of the bindXX methods!
   * 
   */
  public void setBufferEnabled(boolean externalTriggering) {
    this.triggering = externalTriggering;
  }


  //---------------------------accessing
  
  /** set the target to bind.
   *  May be null.
   */
  public void setTarget(Object newTarget) {

    // PENDING: changed from internalflush without testing 
    flush();
    getBeanAdapter().setBean(newTarget);
  }

  /** 
   * returns the bound target. May be null.
   */
  public Object getTarget() {
    return getBeanAdapter().getBean();
  }

  /** 
   * returns a common ancestor of bound components.
   * @return
   */
  public JComponent getInputContainer() {
    return input;
  }

  public boolean isBufferEnabled() {
    return triggering;
  }

  public void commit() {
    internalCommit();
    if (externalTrigger != null) {
      externalTrigger.triggerCommit();
    }
  }

  public void flush() {
    internalFlush();
    if (externalTrigger != null) {
      externalTrigger.triggerFlush();
    }
  }

//---------------------------bind by MetaData

  /** tries to bind components to all given bindingMetaData.
   *
   * @param bindingMetaData - list containing metaData. Must not be null.
   * @throws IllegalStateException if !isBindable()
   */
  public void bind(Collection bindingMetaData) {
    checkBindable();
    for (Iterator iter = bindingMetaData.iterator(); iter.hasNext(); ) {
      BindingMetaData metaData = (BindingMetaData) iter.next();
      bind(metaData);
    }
  }


  //---------------------------binding textfield
  public void bindTextInput(String componentName) {
    bindTextInput(componentName, componentName);
  }

  public void bindTextInput(String componentName, boolean editable) {
    bindTextInput(componentName, componentName, editable);
  }

  public void bindTextInput(String componentName, String propertyName) {
    checkBindable();
    bind(new TextBindingMetaData(componentName, propertyName,
        JTextField.class));
  }

  /** binds a TextField to a property. The editable state will be set as
   *  passed in. <p>
   *
   *
   *  PENDING: currently _not_ adapted to Document directly, that is
   *  inserts/remove will not be committed at once but on
   *  on internal (action and focusLost)
   *  trigger, need parameter to enable/disable on a per-binding basis.
   *
   */
  public void bindTextInput(String componentName, String propertyName,
      boolean editable) {
    checkBindable();
    TextBindingMetaData metaData = new TextBindingMetaData(componentName,
                                     propertyName, JTextField.class);
    metaData.setEditable(editable);
    bind(metaData);
  }

//----------------------------- TextArea
  public void bindTextAreaInput(String componentName) {
    bindTextAreaInput(componentName, componentName);
  }

  public void bindTextAreaInput(String componentName, boolean editable) {
    bindTextAreaInput(componentName, componentName, editable);
  }

  public void bindTextAreaInput(String componentName, String propertyName) {
    checkBindable();
    bind(new TextBindingMetaData(componentName, propertyName,
        JTextArea.class));
  }

  /** binds a TextArea to a property. The editable state will be set as
   *  passed in. <p>
   *
   *  PENDING: currently _not_ adapted to Document directly, that is
   *  inserts/remove will not be committed at once but on
   *  on internal (action and focusLost)
   *  trigger, need parameter to enable/disable on a per-binding basis.
   *
   * @param componentName
   * @param propertyName
   * @param editable
   */
  public void bindTextAreaInput(String componentName, String propertyName,
      boolean editable) {
    checkBindable();
    TextBindingMetaData metaData = new TextBindingMetaData(componentName,
                                     propertyName, JTextArea.class);
    metaData.setEditable(editable);
    bind(metaData);
  }

//--------------------------- binding formatted text
  public void bindFormattedTextInput(String componentName) {
    bindFormattedTextInput(componentName, componentName);
  }

  public void bindFormattedTextInput(String componentName,
      String propertyName) {
    checkBindable();
    BindingMetaData metaData = new FormattedTextBindingMetaData(componentName,
                                 propertyName);
    bind(metaData);

  }
//---------------------------binding spinner
  public void bindSpinnerInput(String componentName) {
    bindSpinnerInput(componentName, componentName);
  }

  public void bindSpinnerInput(String componentName, String propertyName) {
    bindSpinnerInput(componentName, propertyName, null);
  }

  public void bindSpinnerInput(String componentName, SpinnerModel model) {
    bindSpinnerInput(componentName, componentName, model);
  }

  /** binds a spinner's value to the property. Set's the spinnerModel if
   *  != null. Installs a SpinnerConnector with the spinner's model to
   *  guarantee updates of the property. <p>
   *
   * All shortcut methods for spinner binding call this method.
   *
   * @param componentName
   * @param propertyName
   * @param spinnerModel the spinnerModel to use, may be null.
   */
  public void bindSpinnerInput(String componentName, String propertyName,
      SpinnerModel spinnerModel) {
    checkBindable();
    bind(new SpinnerBindingMetaData(componentName, propertyName,
        spinnerModel));

  }
  //---------------------------binding comboBox
  public void bindComboBoxInput(String componentName, Object model) {
    bindComboBoxInput(componentName, componentName, model);
  }

  public void bindComboBoxInput(String componentName, Object model,
      boolean editable) {
    bindComboBoxInput(componentName, componentName, model, editable);
  }

  public void bindComboBoxInput(String componentName, String propertyName,
      Object model) {
    checkBindable();
    ComboBoxBindingMetaData metaData =
      new ComboBoxBindingMetaData(componentName, propertyName, model);
    bind(metaData);
  }

  /** binds the selection of a comboBox to a property. The available
   *  values are passed in model. The editable state is set according
   *  the editable flag. <p>
   *
   *  PRE: model can an Array, a ListModel, a List, or a
   *  property returning a ListModel or a List.
   *  Internally the binding will create a
   *  ComboBoxAdapter for all.  <p>
   *
   *  All short-cut methods for binding combos use this.
   */
  public void bindComboBoxInput(String componentName, String propertyName,
      Object model, boolean editable) {
    checkBindable();
    if (model instanceof String) {
      model = getValueModel((String) model);
    }
    ComboBoxBindingMetaData metaData =
      new ComboBoxBindingMetaData(componentName, propertyName, model);
    metaData.setEditable(editable);
    bind(metaData);
  }

  //---------------------------binding List

  /**
   * binds a list with both data and (single) selection.
   * listPropertyName is the source of the data, 
   * selectionPropertyName is the property to synch with on selection,
   * may be null. <p>
   * 
   * NOTE: clients are responsible to guarantee that listProperty
   * returns a "list-like" structure usable by SelectionInList as
   * listHolder.
   * 
   */
  public void bindListAndSelectionInput(String componentName,
      String listPropertyName, String selectionPropertyName) {
    checkBindable();
    ListBindingMetaData metaData = new ListBindingMetaData(componentName,
                                     listPropertyName);
    if (selectionPropertyName != null) {
      metaData.setSelectionValueModel(getValueModel(selectionPropertyName));
    }
    bind(metaData);
  }
  
  /**
   * binds a list with both (single) selection and data.
   * selectionPropertyName is the property to synch with on selection,
   * listPropertyName is the source of the data. <p>
   *
   * @deprecated: here focus is too strong on "selection" as property -
   *   a JList should be bound to a "list-like" property 
   */
  public void bindListInput(String componentName,
      String selectionPropertyName, String listPropertyName) {
    bindListAndSelectionInput(componentName, listPropertyName, selectionPropertyName);
  }

  /**
   * binds a list with both data and (single) selection.
   * listPropertyName is the source of the data and selectionPropertyName
   * is the property to synch with on selection. <p>
   * @deprecated: this is the very old method which (accidentally?)
   *   got the focus of a JList on representing list-like structure
   *   right. 
   */
  public void bindListAndSelection(String componentName,
      String listPropertyName, String selectionPropertyName) {
    bindListInput(componentName, selectionPropertyName, listPropertyName);
  }

  //---------------------------binding togglebuttons
  public void bindBooleanInput(String componentName) {
    bindBooleanInput(componentName, componentName);
  }

  /** binds an AbstractButton to a two-state property. <p>
   *
   * Internally uses a ToggleButtonAdapter. For constraints on the
   * property, have a look at its api doc.<p>
   *
   * PENDING: client is responsible for not binding a radiobuttons/toggleButtons
   * by this method at least none that should be used in a radioGroup.
   *
   * @param componentName
   * @param propertyName
   */
  public void bindBooleanInput(String componentName, String propertyName) {
    checkBindable();
    BindingMetaData metaData = new BindingMetaData(componentName,
                                 propertyName, AbstractButton.class);
    bind(metaData);
  }


//---------------------------- RadioButtons

  /** binds the selection in a group of RadioButtons to a property.
   *  The first list contains the componentIDs of the buttons, the
   *  second list contains the list of the corresponding values to
   *  set to the property on selection of this button. <p>
   *
   *  NOTE: it's the calling code's responsibility to guarantee valid
   *  options and equal lengths on boths lists. <p>
   *
   *  @deprecated: consistency - sequence of parameters: 
   *  By internal convention the
   *  componentName is first - here we have a list of componentNames...
   */
  public void bindRadioGroup(String propertyName, List componentIDs,
      List optionValues) {
    bindRadioGroup(componentIDs, propertyName, optionValues);
  }

  /** binds the selection in a group of RadioButtons to a property.
   *  The first list contains the componentIDs of the buttons, the
   *  second list contains the list of the corresponding values to
   *  set to the property on selection of this button. <p>
   *
   *  NOTE: it's the calling code's responsibility to guarantee valid
   *  options and equal lengths on boths lists. <p>
   *
   */
  public void bindRadioGroup(List componentIDs, String propertyName, 
      List optionValues) {
    checkBindable();
    RadioBindingMetaData metaData = new RadioBindingMetaData(componentIDs,
                                      propertyName, optionValues);
    bind(metaData);
  }
//---------------------------managing component state

  /** adds a collection of components to be (dis)abled in synch with
   *  the selection state of the AbstractButton componentName.
   *
   * @param componentName the id of the controlling button
   * @param dependents a list of ids of dependent components
   */
  public void addToggleDependents(String componentName, List dependents) {
    addToggleDependents(componentName, dependents, false);
  }

  public void addInversToggleDependents(String componentName,
      List dependents) {
    addToggleDependents(componentName, dependents, true);
  }

  public void setEnabled(String componentName, boolean enabled) {
    JComponent comp = findComponent(componentName);
    if (comp != null) {
      comp.setEnabled(enabled);
    }
  }

  public JComponent findComponent(String name) {
    return KleoUtilities.findJChildByName(getInputContainer(), name, true);
  }

//--------------------- helpers for controlling multiplexing state
  private void addToggleDependents(String componentName, List dependents,
      boolean isInvers) {
    checkBindable();
    EnableMultiplexer multiplexer = getEnableMultiplexer(componentName);
    if (multiplexer == null) {
      return;
    }
    multiplexer.setInvers(isInvers);
    addToggleDependents(multiplexer, dependents);
  }

  /**
   * pre: isBindable()
   * @param componentName
   * @return
   */
  private EnableMultiplexer getEnableMultiplexer(String componentName) {
    EnableMultiplexer multiplexer =
      (EnableMultiplexer) getMultiplexers().get(componentName);
    if (multiplexer == null) {
      AbstractButton button =
        KleoUtilities.findAbstractButton(getInputContainer(), componentName);
      if (button == null) {
        return null;
      }
      multiplexer = new EnableMultiplexer(button);
      getMultiplexers().put(componentName, multiplexer);
    }
    return multiplexer;
  }

  protected void addToggleDependents(EnableMultiplexer multiplexer,
      List dependents) {
    for (Iterator iterator = dependents.iterator(); iterator.hasNext(); ) {
      JComponent comp = findComponent(String.valueOf(iterator.next()));
      if (comp != null) {
        multiplexer.addDependent(comp);
      }
    }
  }

  //---------------------------helper for handling adapting

  /** returns an adapter from the beanAdapter.
   *  will wrap the adapter into a DelayedRefresher if not
   *  observeChanges (to guarantee synch with bean as closely
   *  as possible)
   *
   *  PENDING: incomplete analysis internalTrigger/delayedRefresher/
   *  externaltrigger....
   */
  protected ValueModel getValueModel(String propertyName) {
    ValueModel model = getBeanAdapter().getValueModel(propertyName);
    if (isBufferEnabled()) {
      model = new BufferedValueModel(model, getExternalTrigger());
    }
    if (!observeChanges) {

      // PENDING: needs to be testet!!
      model = new DelayedRefresher(model);
    }
    return model;
  }


  /**
   * PRE: isBindable()
   */
  protected void bind(BindingMetaData metaData) {
    if (metaData.isCommitOnFocusLost()) {
      metaData.setFocusLostTrigger(getInternalTrigger());
    }
    BindingHandler handler = getBindingHandler(metaData);
    handler.bind(getInputContainer(), metaData,
                 getValueModel(metaData.getPropertyName()));
  }


  protected BindingHandler getBindingHandler(BindingMetaData metaData) {
    return BindingFactory.getBindingMap().getBindingHandler(metaData);
  }

  
  /** 
   * not really needed if null check in constructor
   * @return
   */
  protected boolean isBindable() {
    return (getInputContainer() != null);
  }

  /** 
   * should not happen...
   * @throws BindingComponentNotFoundException if !isBindable()
   */
  protected void checkBindable() throws BindingComponentNotFoundException {
    if (!isBindable()) {
      throw new BindingComponentNotFoundException("cannot bind with null input container");
    }
  }

  /** will trigger a commit from internal events.
   *  (f.i. actionEvents on textfields, or focusLost).
   *
   */
  protected void internalCommit() {
    getInternalTrigger().triggerCommit();
  }

  protected void internalFlush() {
    getInternalTrigger().triggerFlush();
  }

  //---------------------------factory methods - lazy
  
  /** returns the external trigger.
   *  creates it if null.
   */
  protected Trigger getExternalTrigger() {
    if (externalTrigger == null) {
      externalTrigger = new Trigger();
    }
    return externalTrigger;
  }

  protected BeanAdapter getBeanAdapter() {
    if (beanAdapter == null) {
      beanAdapter = createBeanAdapter();
    }
    return beanAdapter;
  }

  protected BeanAdapter createBeanAdapter() {
    return new BeanAdapter((Object) null, observeChanges);
  }
  private Map getMultiplexers() {
    if (multiplexers == null) {
      multiplexers = createMap();
    }
    return multiplexers;
  }

  private Trigger getInternalTrigger() {
    if (internalTrigger == null) {
      internalTrigger = new Trigger();
    }
    return internalTrigger;
  }

  //---------------------------factory methods
  private Object[] createArray(ListModel model) {
    Object[] values = new Object[model.getSize()];
    for (int i = 0; i < values.length; i++) {
      values[i] = model.getElementAt(i);
    }
    return values;
  }

  protected Map createMap() {
    return new HashMap();
  }

  protected List createList() {
    return new ArrayList();
  }
}
