/*
 * Created on 23.09.2004
 *
 */
package de.kleopatra.view.binding;

import java.awt.event.*;
import java.util.ArrayList;
import java.util.List;

import javax.swing.*;
import javax.swing.text.JTextComponent;

import com.jgoodies.binding.adapter.*;
import com.jgoodies.binding.list.SelectionInList;
import com.jgoodies.binding.value.*;

import de.kleopatra.support.HandlerMap;
import de.kleopatra.support.KleoUtilities;

/**
 * Default implementation of BindingMap.<p>
 *
 * Provides and maps default BindingHandlers for the most common
 * binding types. Most handlers delegate the actual binding
 * by calling Bindings.bind(...) in JGoodies Binding.
 *
 * PENDING: support all standard comps!
 *
 * @author (C) 2004 Jeanette Winzenburg, Berlin
 * @version $Revision: 1.6 $ - $Date: 2004/10/04 11:56:04 $
 */
public class DefaultBindingMap extends HandlerMap implements BindingMap {

  public DefaultBindingMap() {
    initBindingHandlers();
  }

  //------------------------ BindingMap 

  /** Returns a BindingHandler for metaData. <p>
   *  Lookup strategy:
   *  <ul>
   *  <li> BindingMetaData class </li>
   *  <li> component class as returned by metaData </li>
   *  <li> target class as returned by metaData </li>
   *  </ul>
   *
   * @throws BindingHandlerNotFoundException if no appropriate
   *  handler is found.
   */
  public BindingHandler getBindingHandler(BindingMetaData metaData) {
    Object handler = getHandler(metaData.getClass());
    if (handler == null) {
      handler = getHandler(metaData.getComponentClass());
    }
    if (handler == null) {
      handler = getHandler(metaData.getTargetClass());
    }
    if (handler == null) {
      throw new BindingHandlerNotFoundException(metaData);
    }
    return (BindingHandler) handler;
  }

  public void addBindingHandler(Class handledClass, Class handlerClass) {
    addHandler(handledClass, handlerClass);
  }

  public void addBindingHandler(String handledClassName,
      String handlerClassName) {
    addHandler(handledClassName, handlerClassName);
  }

  //--------------------- helper
  protected boolean checkType(Object handler) {
    return handler instanceof BindingHandler;
  }

  protected void initBindingHandlers() {
    handlers = createMap();
    addBindingHandler(JFormattedTextField.class,
        DefaultBindingMap.FormattedTextBindingHandler.class);
    addBindingHandler(FormattedTextBindingMetaData.class,
        DefaultBindingMap.FormattedTextBindingHandler.class);
    addBindingHandler(JSpinner.class,
        DefaultBindingMap.SpinnerBindingHandler.class);
    addBindingHandler(SpinnerBindingMetaData.class,
        DefaultBindingMap.SpinnerBindingHandler.class);
    addBindingHandler(TextBindingMetaData.class,
        DefaultBindingMap.TextBindingHandler.class);
    addBindingHandler(JTextComponent.class,
        DefaultBindingMap.TextBindingHandler.class);
    addBindingHandler(ComboBoxBindingMetaData.class,
        DefaultBindingMap.ComboBoxBindingHandler.class);
    addBindingHandler(JComboBox.class,
        DefaultBindingMap.ComboBoxBindingHandler.class);
    addBindingHandler(ListBindingMetaData.class,
        DefaultBindingMap.ListBindingHandler.class);
    addBindingHandler(AbstractButton.class,
        DefaultBindingMap.BooleanBindingHandler.class);
    addBindingHandler(RadioBindingMetaData.class,
        DefaultBindingMap.RadioBindingHandler.class);
  }

//----------------------------- BindingHandlers
//-------------------- Group of RadioButtons

  /**
   *  Handler for RadioButtons.
   */
  public static class RadioBindingHandler implements BindingHandler {

    /**
     * binds a group of radioButtons to valueModel. Each button represents
     * one of the optionValues provided by metaData. <p>
     *
     * PENDING: extend to include all radio-type buttons (like radioMenuItems)
     */
    public void bind(JComponent container, BindingMetaData metaData,
        ValueModel model) {

      // PENDING: need to fire exception?
      if (!(metaData instanceof RadioBindingMetaData)) {
        return;
      }
      RadioBindingMetaData radioMetaData = (RadioBindingMetaData) metaData;
      List radioButtons = findButtons(container, radioMetaData);
      for (int i = 0; i < radioButtons.size(); i++) {
        JRadioButton radio = (JRadioButton) radioButtons.get(i);
        Bindings.bind(radio, model, radioMetaData.getOptionValue(i));

        //     radio.setModel(new RadioButtonAdapter(model, radioMetaData.getOptionValue(i)));
      }
    }

    private List findButtons(JComponent container,
        RadioBindingMetaData metaData) {
      List buttons = new ArrayList();
      for (int i = 0; i < metaData.getGroupSize(); i++) {
        JRadioButton radio = KleoUtilities.findRadioButton(container,
                               metaData.getComponentID(i));
        if (radio == null) {
          throw new BindingComponentNotFoundException(container, metaData);
        }
        buttons.add(radio);
      }
      return buttons;
    }
  }

//-------------------------- AbstractButton 

  /**
   *  Handler for AbstractButtons (CheckBox-like).
   *
   */
  public static class BooleanBindingHandler implements BindingHandler {

    /**
     * binds an AbstractButton to the valueModel. <p>
     * Assumes that the valueModel has to return a boolean value.<p>
     * PENDING: does not check for radio button, currently client
     *   is responsible for not accidentally finding a radio-type
     *   or action-type button.
     */
    public void bind(JComponent container, BindingMetaData metaData,
        ValueModel model) {
      AbstractButton comp = KleoUtilities.findAbstractButton(container,
                              metaData.getComponentName());
      if (comp == null) {
        throw new BindingComponentNotFoundException(container, metaData);
      }
      boolean enabled = comp.isEnabled();
      comp.setModel(new ToggleButtonAdapter(model));
      comp.setEnabled(enabled);
    }
  }

//------------------ List

  /**
   *  Handler for Lists.
   *
   */
  public static class ListBindingHandler implements BindingHandler {

    /**
     * Creates a SelectionInList out of listHolder (and selectionValueModel, if
     * provided by the metaData). The JList is bound to the sil
     * using the SelectionInList's list model as list data provider
     * and the SelectionInList's selection index holder for the selection model. <p>
     *
     * NOTE: the listHolder must vend a list-like value as accepted by a
     * SelectionInList.
     */
    public void bind(JComponent container, BindingMetaData metaData,
        ValueModel listHolder) {
      JList list = KleoUtilities.findList(container,
                     metaData.getComponentName());
      if (list == null) {
        throw new BindingComponentNotFoundException(container, metaData);
      }
      ValueModel selectionModel = null;
      if (metaData instanceof ListBindingMetaData) {
        selectionModel =
          ((ListBindingMetaData) metaData).getSelectionValueModel();
      }
      SelectionInList sil = createSelectionInList(listHolder, selectionModel);
      Bindings.bind(list, sil);

//      list.setModel(sil);
//      list.setSelectionModel(new SingleListSelectionAdapter(sil
//          .getSelectionIndexHolder()));
    }

    private SelectionInList createSelectionInList(ValueModel listHolder,
        ValueModel selectionHolder) {
      if (selectionHolder == null) {
        return new SelectionInList(listHolder);
      }
      return new SelectionInList(listHolder, selectionHolder);
    }
  }

//-------------------------- ComboBox  

  /**
   *  Handler for ComboBox.
   *
   */
  public static class ComboBoxBindingHandler implements BindingHandler {

    public void bind(JComponent container, BindingMetaData metaData,
        ValueModel model) {
      JComboBox box = KleoUtilities.findComboBox(container,
                        metaData.getComponentName());
      if (box == null) {
        throw new BindingComponentNotFoundException(container, metaData);
      }
      Object comboModel = null;
      if (metaData instanceof ComboBoxBindingMetaData) {
        comboModel = ((ComboBoxBindingMetaData) metaData).getModel();
      }
      if (model == null) {
        comboModel = box.getModel();
      }
      box.setModel(createComboBoxModel(model, comboModel));
      if (metaData instanceof ComboBoxBindingMetaData) {
        box.setEditable(((ComboBoxBindingMetaData) metaData).isEditable());
      }
    }

    private ComboBoxModel createComboBoxModel(ValueModel valueModel,
        Object comboModel) {
      if (comboModel instanceof ListModel) {
        return new ComboBoxAdapter((ListModel) comboModel, valueModel);
      } else if (comboModel instanceof List) {
        return new ComboBoxAdapter((List) comboModel, valueModel);
      } else if (comboModel instanceof Object[]) {
        return new ComboBoxAdapter((Object[]) comboModel, valueModel);
      } else if (comboModel instanceof ValueModel) {
        return new ComboBoxAdapter((ValueModel) comboModel, valueModel);
      }
      throw new IllegalStateException(
        "type mismatch - should not have happened");
    }
  }

//---------------------- TextComponents  
  public static final Object CANCEL_ACTION_KEY = "cancelAction";
  public static final String CANCEL_KEY_STROKE = "ESCAPE";

  /**
   *  Handler for TextComponents (TextFields and TextAreas).
   *
   */
  public static class TextBindingHandler implements BindingHandler {

    /**
     * binds text component to model by replacing the document with a
     * DocumentAdapter. <p>
     * Buffer the commit with metadata's focusLostTrigger if provided.
     * Commits on enter action.
     *
     * PENDING: will not work with JTextPane/JEditorPane because
     * DocumentAdapter is a PlainDocument.
     */
    public void bind(JComponent container, BindingMetaData metaData,
        ValueModel model) {
      JTextComponent textComponent =
        KleoUtilities.findTextComponent(container,
          metaData.getComponentName());
      if (textComponent == null) {
        throw new BindingComponentNotFoundException(container, metaData);
      }
      bindTextComponent(textComponent, metaData, model);
    }

    protected void bindTextComponent(JTextComponent textComponent,
        BindingMetaData metaData, ValueModel model) {
      if (metaData.getFocusLostTrigger() != null) {
        model = new BufferedValueModel(model, metaData.getFocusLostTrigger());
      }
      textComponent.setDocument(new DocumentAdapter(model,
          textComponent.getDocument()));
      if (metaData instanceof TextBindingMetaData) {
        textComponent
          .setEditable(((TextBindingMetaData) metaData).isEditable());
      }
      configure(textComponent, metaData);
    }

    protected void configure(JTextComponent textComponent,
        BindingMetaData metaData) {
      if (metaData.getFocusLostTrigger() == null) {
        return;
      }
      ActionFocusHandler handler =
        new ActionFocusHandler(metaData.getFocusLostTrigger());
      textComponent.addFocusListener(handler);
      if (textComponent instanceof JTextField) {

        // PENDING ... should exclude JFormattedTextFields explicitly -
        // to be on the safe side??
        // usually they are not handled here because 
        // any decent BindingMap will map them to a dedicated
        // FormattedBindingHandler...
        ((JTextField) textComponent).addActionListener(handler);
      }

      // PENDING ... expect trouble in a commit/reset scheme??
//      textComponent.getInputMap().put(
//          KeyStroke.getKeyStroke(CANCEL_KEY_STROKE), CANCEL_ACTION_KEY);
//      textComponent.getActionMap().put(CANCEL_ACTION_KEY,
//          new TriggerCancelAction(metaData.getFocusLostTrigger()));
    }
  }

  private static class ActionFocusHandler extends FocusAdapter
      implements ActionListener {

    private Trigger trigger;
    public ActionFocusHandler(Trigger trigger) {
      this.trigger = trigger;
    }

    public void focusLost(FocusEvent e) {
      if (!e.isTemporary()) {
        actionPerformed(null);
      }
    }

    public void actionPerformed(ActionEvent e) {
      trigger.triggerCommit();
    }
  }

  /**
   * action to flush trigger, typically bound to cancel keystroke.<p>
   * The idea is to easily undo an edit in commitOnFocusLost scenarios.
   * 
   *  PENDING: problems with flushing unrelated when chaining to 
   *  buffered valueModels (like in dialogs). 
   */
  public static class TriggerCancelAction extends AbstractAction {

    private Trigger trigger;
    public TriggerCancelAction(Trigger trigger) {
      this.trigger = trigger;
    }

    public void actionPerformed(ActionEvent e) {
      trigger.triggerFlush();
    }
  }

//--------------------- Spinner  

  /**
   *  .
   *	<p>
   *
   * @author (C) 2004 Jeanette Winzenburg, Berlin
   * @version $Revision: 1.6 $ - $Date: 2004/10/04 11:56:04 $
   *
   */
  public static class SpinnerBindingHandler implements BindingHandler {

    /**
     * Connects model to JSpinner's model. <p>
     * If metaData carries a custom spinnerModel set it
     * to the found JSpinner before connecting.
     */
    public void bind(JComponent container, BindingMetaData metaData,
        ValueModel model) {
      JSpinner spinner = KleoUtilities.findSpinner(container,
                           metaData.getComponentName());
      if (spinner == null) {
        throw new BindingComponentNotFoundException(container, metaData);
      }
      SpinnerModel spinnerModel = null;
      if (metaData instanceof SpinnerBindingMetaData) {
        spinnerModel = ((SpinnerBindingMetaData) metaData).getModel();
      }
      if (spinnerModel == null) {
        spinnerModel = spinner.getModel();
      } else {
        spinner.setModel(spinnerModel);
      }
      new SpinnerConnector(spinnerModel, model);
    }
  }

//---------------------- FormattedText

  /**
   *  .
   *	<p>
   *
   * @author (C) 2004 Jeanette Winzenburg, Berlin
   * @version $Revision: 1.6 $ - $Date: 2004/10/04 11:56:04 $
   *
   */
  public static class FormattedTextBindingHandler implements BindingHandler {

    /**
     *  PENDING: support for different formats/masks.
     */
    public void bind(JComponent container, BindingMetaData metaData,
        ValueModel model) {
      JFormattedTextField formatted =
        KleoUtilities.findFormattedText(container,
          metaData.getComponentName());
      if (formatted == null) {
        throw new BindingComponentNotFoundException(container, metaData);
      }
      Bindings.bind(formatted, model);
    }
  }
}
