package de.kleopatra.view.factory;

import java.util.*;
import java.util.logging.Logger;

import de.kleopatra.persistence.ResourceManager;

/**	Default implementation.
 * <p>
 *	Handles storage of classnames/classes, 
 *  delegates classloading to ResourceManager,
 *  instantiates comps,
 *  provides hook for custom factories.
 *  <p>
 *  NOTE: the custom hook is processed after looking for
 *  a mapped class. This guarantees that type mappings added
 *  at runtime take precedence over hardcoded lookups.
 * 
 * @author (C) Jeanette Winzenburg, Berlin
 * @version $Revision: 1.1.1.1 $ - $Date: 2004/06/22 12:32:37 $
 *
 */
public class DefaultComponentTypeMap implements ComponentTypeMap {

	protected Map componentMap;
	protected Map containerMap;
	protected Set scrollableTypes;
	protected static final Logger LOGGER = Logger.getLogger("DefaultComponentTypeMap");
	
	public DefaultComponentTypeMap() {
	}

//-----------------ComponentTypeMap
	
	public void addComponentType(Object componentType, String className) {
		getComponentMap().put(componentType, className);
	}

	public void addContainerType(Object containerType, String className) {
		getContainerMap().put(containerType, className);
	}

	public void addScrollableType(Object scrollableType) {
		getScrollableTypes().add(scrollableType);
	}

	public boolean shouldScroll(Object componentType) {
		return getScrollableTypes().contains(componentType);

	}

//-----------------------------ExComponentTypeMap
	public void addComponentType(Object componentType, Class clazz) {
		getComponentMap().put(componentType, clazz);
	}

	public void addContainerType(Object componentType, Class clazz) {
		getContainerMap().put(componentType, clazz);
	}

	/** 
	 *  returns an instance of a component as appropriate for
	 *  the generic type componentType.
	 *  <p>
	 *  Implementation note: guarantees to first check the registered
	 *  class mapping before trying to use thefactoryHook
	 */
	public Object getComponentInstance(Object componentType) {
		Class clazz = getClass(componentType);
		if (clazz != null) {
			try {
				return clazz.newInstance();
			} catch (Exception ex) {
				LOGGER.warning("Instantiation failed - can't do anything: "+ ex);
			}
		}
		return factoryHook(componentType);
	}

	public Set getComponentTypes() {
		return Collections.unmodifiableSet(getComponentMap().keySet());
	}

	public Set getContainerTypes() {
		return Collections.unmodifiableSet(getContainerMap().keySet());
	}

	/** returns an identifier for the given class or
	 *	null if not found.
	 *  used by life form readers.
	 * 
	 * 	PENDING: does not yet cope with classes mapped directly 
	 *  (as opposed to mapped via className)
	 */
	public Object getComponentType(Class clazz) {
		String name = clazz.getName();
		Set entries = getComponentMap().entrySet();
		for (Iterator iterator = entries.iterator(); iterator.hasNext(); ) {
			Map.Entry entry = (Map.Entry) iterator.next();
			if (entry.getValue().equals(name)) {
				return entry.getKey();
			}
		}
		entries = getContainerMap().entrySet();
		for (Iterator iterator = entries.iterator(); iterator.hasNext(); ) {
			Map.Entry entry = (Map.Entry) iterator.next();
			if (entry.getValue().equals(name)) {
				return entry.getKey();
			}
		}
		return null;
	}

//-----------------------------Helper
	
	/** returns a class for a given componentTyp
	 *	null if not found or error.
	 *  used by UIBuilder.
	 */
	protected Class getClass(Object componentType) {
		try {
			Object classEntry = getComponentMap().get(componentType);
			if (classEntry != null) {
				return getClassForEntry(classEntry);
			}
			classEntry = getContainerMap().get(componentType);
			if (classEntry != null) {
				return getClassForEntry(classEntry);
			}
		} catch (ClassNotFoundException ex) {
			LOGGER.warning("Unknown Class - can't do anything about: "+ ex);
		}
		return null;
	}

	protected Class getClassForEntry(Object classEntry) throws ClassNotFoundException {
		if (classEntry instanceof Class) {
			return (Class) classEntry;
		}
		if (classEntry instanceof String) {
			return ResourceManager.getInstance().getClass((String) classEntry);
		}
		return null;
	}

	
	/** hook for subclasses to use custom component factories. 
	 *  Should return a component instance for the generic type
	 *  or null.
	 * 
	 *  here: returns null; 
	 */
	protected Object factoryHook(Object componentType) {
		return null;
	}
	
//---------------------------factory methods


	protected Map getComponentMap() {
		if (componentMap == null) {
			componentMap = createMap();
		}
		return componentMap;
	}

	protected Map getContainerMap() {
		if (containerMap == null) {
			containerMap = createMap();
		}
		return containerMap;
	}


	protected Map createMap() {
		return new HashMap();
	}

	protected Collection getScrollableTypes() {
		if (scrollableTypes == null) {
			scrollableTypes = createSet();
		}
		return scrollableTypes;
	}

	protected Set createSet() {
		return new HashSet();
	}


}
