package de.kleopatra.view.factory;

import java.awt.*;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Logger;

import javax.swing.*;
import javax.swing.event.ChangeListener;

import de.kleopatra.persistence.ResourceManager;


/**	Convenience factory for divers classes.
 *
 *	Mainly to ease component creation/customization,
 *	deals with buggy comps
 *
 *	<p> Note: has nothing to do with UIFactory, but
 *	think: how to integrate with UIFactory?
 *	this here is  special for Swing
 *
 *	<p>
 *	creates a SwingGoodiesFactory if UIManager has a boolean.true for key "Framework.JGoodies"
 *
 * @author (C) Jeanette Winzenburg, Berlin
 * @version $Revision: 1.1.1.1 $ - $Date: 2004/06/29 12:43:01 $
 *
 */
public class SwingFactory {
	private static final Insets EMPTY_INSETS = new Insets(0, 0, 0, 0);
	
	private static final Logger LOGGER = Logger.getLogger("SwingFactory");
	private static SwingFactory factory;
	private ChangeListener tabbedPaneListener;

	protected SwingFactory() {
	}

	public static SwingFactory getInstance() {
		if (factory == null) {
			factory = createFactory();
		}
		return factory;
	}

//---------------------------menus

	public JMenuBar createMenuBar() {
		return new JMenuBar();
	}

	public JMenu createMenu(Object id) {
		String menuText = String.valueOf(id);
		char acc = Character.toUpperCase(menuText.charAt(0));
		JMenu menu = new JMenu(String.valueOf(id));
		menu.setMnemonic(acc);
		return menu;
	}

	public JMenuItem createMenuItem(Action action) {
		JMenuItem item = new JMenuItem(action) {
			protected void configurePropertiesFromAction(Action a) {
				super.configurePropertiesFromAction(a);
				if (Boolean.TRUE.equals(UIManager.get("Menu.ignoreIcon"))) {
					setIcon(null);
				}
			}
		};
		return item;

	}

	public JMenuItem createRadioMenuItem(Action action) {
		JMenuItem item = new JRadioButtonMenuItem(action) {
			protected void configurePropertiesFromAction(Action a) {
				super.configurePropertiesFromAction(a);
				if (Boolean.TRUE.equals(UIManager.get("Menu.ignoreIcon"))) {
					setIcon(null);
				}
			}
		};
		return item;
	}

	public void addSeparator(JMenu menu) {
		menu.addSeparator();
	}
	
//---------------------------buttons + toolbars

	public AbstractButton createButton(Action action, boolean toggle) {
		if (toggle) {
			return new JToggleButton(action);
		}
		return new JButton(action);
	}

	public AbstractButton createToolBarButton(Action action, boolean toggle) {
		AbstractButton button = createButton(action, toggle);
		if (button.getIcon() != null) {
			// hmmm... should be enough...
			button.putClientProperty("hideActionText", Boolean.TRUE);
			// unreliable hack: button's propertyChangeListener does not update correctly on
			// setting the property to true (does nothing)....
			button.setText(null);
		}
		button.setHorizontalTextPosition(JButton.CENTER);
		button.setVerticalTextPosition(JButton.BOTTOM);
		button.setRequestFocusEnabled(false);
		button.setFocusable(false);
		return button;
	}

	public JComponent buildToolPanel(List actions, boolean toggle, int columns) {
		JToolBar toolBar = createToolBar(JToolBar.VERTICAL);
		ButtonGroup group = new ButtonGroup();
		toolBar.setLayout(new GridLayout(0, columns));
		boolean hasSelected = false;
		for (Iterator iterator = actions.iterator(); iterator.hasNext();) {
			AbstractButton button = createToolBarButton((Action) iterator.next(), toggle);
			toolBar.add(button);
			if (toggle) {
				group.add(button);
				if ((!hasSelected) && (button.isEnabled())) {
				//hmm.... assumes that the action is valid at this point
					button.doClick();
					hasSelected = true;
				}
			}
		}
		toolBar.setMaximumSize(toolBar.getPreferredSize());
		JComponent comp = Box.createVerticalBox();
		comp.add(toolBar);
		return comp;
	}

	public JToolBar buildToolBar(List actions, int orientation) {
		JToolBar toolBar = createToolBar(orientation);
		addToToolBar(toolBar, actions, false);
		return toolBar;
	}

	public void addToToolBar(JToolBar toolBar, List actions, boolean withSeparator) {
	  if (actions == null) return;
		if (withSeparator) {
			addSeparator(toolBar);
		}
		for (Iterator iterator = actions.iterator(); iterator.hasNext();) {
			toolBar.add(createToolBarButton((Action)iterator.next(), false));
		}
	}

	public void addSeparator(JToolBar toolBar) {
		toolBar.addSeparator();
	}

	public JToolBar createToolBar() {
		JToolBar toolbar = createToolBar(JToolBar.HORIZONTAL);
		toolbar.setFocusable(false);
		return toolbar;

	}

	public JToolBar createToolBarAndButtons(List actions, UIFactory factory) {
	  return createToolBarAndButtons(actions, factory, JToolBar.HORIZONTAL);
	}
	public JToolBar createToolBarAndButtons(List actions, UIFactory factory, int direction) {
		JToolBar toolBar = createToolBar(direction);
		addToToolBar(toolBar, actions, factory, false);
		return toolBar;
	}
	
	public void addToToolBar(JToolBar toolBar, List actions, UIFactory factory, boolean withSeparator) {
		if (actions == null) return; 
		if (withSeparator) {
			addSeparator(toolBar);
		}
		for (Iterator iterator = actions.iterator(); iterator.hasNext();) {
			Action element = (Action) iterator.next();
			Component button = factory.createComponent(null, "Button");
			if (button instanceof AbstractButton) {
				((AbstractButton) button).setAction(element);
				button.setName(((AbstractButton) button).getActionCommand());
				//button.setRequestFocusEnabled(false);
				button.setFocusable(false);
			}
			toolBar.add(button);
		}
	}
	
	public void configureToolBarAndButtons(JToolBar toolBar) {
		configureButtons(toolBar);
		configureToolBar(toolBar);
	}
	
	public void configureButtons(JToolBar toolBar) {
		for (int i = 0; i < toolBar.getComponentCount(); i++) {
			if (toolBar.getComponent(i) instanceof AbstractButton) {
				configureToolBarButton((AbstractButton) toolBar.getComponent(i));
			}
		}
	}
	
	public void configureToolBarButton(AbstractButton button) {
		button.setHorizontalTextPosition(JButton.CENTER);
		button.setVerticalTextPosition(JButton.BOTTOM);
		button.setAlignmentY(JComponent.CENTER_ALIGNMENT);
		button.setRequestFocusEnabled(false);
		button.putClientProperty("hideActionText", Boolean.TRUE);
		button.setText(null);
		button.setMargin(getButtonMargin(button));
	}

	protected Insets getButtonMargin(AbstractButton button) {
			Dimension defaultIconSize =
					UIManager.getDimension("jgoodies.defaultIconSize");
			Icon icon = button.getIcon();
			if (defaultIconSize == null || icon == null)
					return EMPTY_INSETS;

			int hpad = Math.max(0, defaultIconSize.width - icon.getIconWidth());
			int vpad = Math.max(0, defaultIconSize.height - icon.getIconHeight());
			int top    = vpad / 2;
			int left   = hpad / 2;
			int bottom = top  + vpad % 2;
			int right  = left + hpad % 2;
			return new Insets(top, left, bottom, right);
	}

	public void configureToolBar(JComponent toolbar){
		toolbar.setFocusable(false);
	}

//----------------------
	public void configureThinBorder(JComponent comp) {
	}
	

	public JScrollPane createScrollPane() {
		JScrollPane scroll = new JScrollPane();
		configureScrollPane(scroll);
		return scroll;
	}
	
	public void configureScrollPane(JScrollPane scroll) {
		scroll.getHorizontalScrollBar().setFocusable(false);
		scroll.getVerticalScrollBar().setFocusable(false);
	}
	
	public JComponent createTablePanel(JTable table) {
		JScrollPane scroll = createScrollPane();
		scroll.setViewportView(table);
		return scroll;
	}

	public JTabbedPane createTabbedPane() {
		JTabbedPane pane = new JTabbedPane();
		pane.putClientProperty("jgoodies.noContentBorder", Boolean.TRUE);
		return pane;
	}
	
//---------------------------factory methods

	protected JToolBar createToolBar(int direction) {
		return new JToolBar(direction);
	}

	private static SwingFactory createFactory() {
		// PENDING: change to store SwingFactory class directly
		Object customFactoryClass = UIManager.get("SwingFactory.class");
		if (customFactoryClass instanceof Class) {
			try {
				SwingFactory factory = (SwingFactory) ((Class) customFactoryClass).newInstance();
				LOGGER.info("Setting custom SwingFactory " + factory.getClass());
				return factory;
			} catch (Exception e) {
				// can't do anything...
				LOGGER.info("could not find/access/instantiate custom SwingFactory " + customFactoryClass
				+ ". Will use default");
			}
		} else { // for backward compatibility only
			Object useGoodies = UIManager.get("Framework.JGoodies");
			if (Boolean.TRUE.equals(useGoodies)) {
				try {
					SwingFactory factory = (SwingFactory) ResourceManager.getInstance().getClass("de.kleopatra.view.factory.SwingGoodiesFactory").newInstance();
					LOGGER.info("setting SwingGoodiesFactory");
					return factory;
				} catch (Exception ex) {
					// can't do anything...
					LOGGER.info("Could not find/access/instantiate SwingGoodiesFactory" 
							+ ". Will use default");
				}
			}
		}

		return new SwingFactory();
	}

  public JPopupMenu createPopupMenu() {
    JPopupMenu popup = new JPopupMenu();
    return popup;
  }

  public JSplitPane createSplitPane() {
    JSplitPane pane = new JSplitPane();
    // TODO Auto-generated method stub
    return pane;
  }



}
