package de.kleopatra.view.factory;

import java.awt.Component;
import java.awt.Container;
import java.lang.reflect.Method;
import java.util.*;

import javax.swing.*;

import de.kleopatra.view.presentation.DefaultPresentationManager;
import de.kleopatra.view.presentation.PresentationManager;

/**	Implementation of UIFactory for Swing toolkit.
 *
 *
 * @author (C) Jeanette Winzenburg, Berlin
 * @version $Revision: 1.1.1.1 $ - $Date: 2004/05/05 10:02:18 $
 * 
 */
public class SwingUIFactory implements UIFactory {

	private ComponentTypeMap componentTypes;
//	private List scrollableTypes;

	private PresentationManager presentationManager;

	public SwingUIFactory() {
		installDefaultComponentTypes();
	}

//---------------------------UIFactory: basics

	public Component createComponent(Object componentID, Object componentType) {
		Component result = createAndCustomizeComponent(componentID, componentType);
		if (result != null) {
			return result;
		}
		return createAndCustomizeComponent(componentID, getFallBackComponentType());
	}

	public void setComponentTypeMap(ComponentTypeMap componentTypeMap) {
		this.componentTypes = componentTypeMap;
	}

	public ComponentTypeMap getComponentTypeMap() {
		return componentTypes;
	}

	public Container createDefaultContainer(Object componentID) {
		return (Container) createComponent(componentID, getDefaultContainerType());
	}

//---------------------------uiFactory: experimental

	public PresentationManager getPresentationManager() {
		if (presentationManager == null) {
			presentationManager = createPresentationManager();
		}
		return presentationManager;
	}

	protected PresentationManager createPresentationManager() {
		return new DefaultPresentationManager();
	}

	/** tries to extract the componentID from a given comp.
	 *	result may be null.
	 *	PRE: comp != null
	 *
	 */
	public Object getComponentID(Component comp) {
		if (comp.getName() != null) {
			return comp.getName();
		}
		if (comp instanceof JComponent) {
			return guessComponentID((JComponent) comp);
		}
		return null;
	}

//---------------------------specifics

	protected Component createAndCustomizeComponent(Object componentID, Object componentType) {
		if (componentTypes != null) {
				Object comp =  componentTypes.getComponentInstance(componentType);
				if (comp instanceof Component) {
					return customizeComponent((Component) comp, componentID, componentType);
				}
		} 

		return null;
	}

	protected Object guessComponentID(JComponent comp) {
		if (comp instanceof AbstractButton) {
			return guessAbstractButtonID((AbstractButton) comp);
		}
		return guessDefaultComponentID(comp);
	}

	protected Component customizeComponent(Component comp, Object componentID, Object componentType) {
	 	comp.setName(String.valueOf(componentID));
	 	if (shouldScroll(componentType)) {
	 		comp = createScroller(comp, componentID);
	 	}
	 	return comp;
	}

	protected Component createScroller(Component component, Object componentID) {
		Component scroller = createComponent(createScrollID(componentID), getScrollPaneType());
		if (scroller instanceof JScrollPane) {
			((JScrollPane) scroller).setViewportView(component);
			SwingFactory.getInstance().configureScrollPane((JScrollPane) scroller);
			return scroller;
		}
		return component;
	}

	protected boolean shouldScroll(Object componentType) {
		return getComponentTypeMap().shouldScroll(componentType);
	}

	protected Object guessAbstractButtonID(AbstractButton comp) {
		Action action = comp.getAction();
		if ((action != null) && (action.getValue(Action.ACTION_COMMAND_KEY) != null)) {
			return action.getValue(Action.ACTION_COMMAND_KEY);
		}
		return guessDefaultComponentID(comp);
	}

	/** fall back to see at least something if presentationManager is not active.
	 *  or exist comps with same names.
	 *	useful in reading given containers (only?)
	 */
/*	protected void customizeComponentDefaultProperty(Component comp, Object componentID) {
		Class clazz = comp.getClass();
		String methodName = "setText";
		Class[] parameters = new Class[] { String.class };
		try {
			Method method = clazz.getMethod(methodName, parameters);
			method.invoke(comp, new Object[] { String.valueOf(componentID) });
		} catch (Exception ex) {
			D.ebug("Error in method invoke - can't do anything... ", ex);
		}
	}
*/
	protected Object guessDefaultComponentID(Component comp) {
		Class clazz = comp.getClass();
		String methodName = "getText";
		try {
			Method method = clazz.getMethod(methodName, null);
			String text = (String) method.invoke(comp, null);
			if ((text != null) && (text.length() == 0)) {
				return null;
			}
			return text;
		} catch (Exception ex) {
			//System.err.println("Error in method invoke - can't do anything... "+ex);
		}
		return null;
	}

//---------------------------helper

	protected Object createScrollID(Object componentID) {
		return String.valueOf(componentID); // + "Scroller";
	}

	protected void installDefaultComponentTypes() {
		componentTypes = createComponentTypeMap();
	}

	protected Object getFallBackComponentType() {
		return "PlaceHolder";
	}

	protected Object getDefaultContainerType() {
		return "Panel";
	}

	protected Object getScrollPaneType() {
		return "ScrollPane";
	}

//---------------------------factory methods - lazy instantiation

//---------------------------factory methods

	protected ComponentTypeMap createComponentTypeMap() {
		return new SwingComponentTypeMap();
	}

	protected Map createMap() {
		return new HashMap();
	}

	protected List createList() {
		return new ArrayList();
	}


}
