package de.kleopatra.view.presentation;

import java.awt.event.ActionListener;
import java.beans.EventHandler;
import java.util.*;

import javax.swing.Action;

/** 
 * Creates, "registers" and wires actions to resourceIDs and eventHandlers.
 * <p>
 * The actions are identifiable by a actionID/resourceID pair.
 * ActionIDs should be unique per resourceID, resourceIDs should be unique
 * per application (nothing done yet to enforce it).
 * Generally, there is one eventHandler per resourceID. 
 * It's possible to replace the handler of a subset of a resourceID, though.
 * 
 * 
 * <p>
 *  PENDING: unique actionID per listID - who is responsible for checking?
 *
 * @author (C) 2003, 2004 Jeanette Winzenburg, Berlin
 * @version $Revision: 1.2 $ - $Date: 2004/09/29 14:40:19 $
 *
 */
public class ActionFactory {

	private static ActionFactory actionFactory;

	// map that holds lists of actions;
	private Map actionLists;
	protected ActionFactory() {
	}

	public static ActionFactory getInstance() {
		if (actionFactory == null) {
			actionFactory = createActionFactory();
		}
		return actionFactory;
	}

	/** used for unittesting only: to get a new singleton.
	 */
	public void reset() {
		actionFactory = createActionFactory();
	}

	/** removes all action/resourceID mappings.
	 * 
	 */
	public void clear() {
		getActionLists().clear();
	}

	/** not used and wrong name: should be hasResourceID(...).
	 * @deprecated
	 */
	public boolean isUniqueListID(Object resourceID) {
		return getActions(resourceID) == null;
	}

	/** 
	 *  registers action with given ids to resource and wires them to 
	 *  given eventhandler.
	 * 
	 * 	<p>
	 *  creates an action if there is no action with the id in the resource map.
	 * 	if there already is an action in the resource map with the id, its 
	 *  eventhandler is replaced by the new one. 
	 * 
	 *  @deprecated use registerActionsAndHandler
	 */
	public List setActionList(List actionIDs, Object resourceID, Object handler) {
		return registerHandler(actionIDs, resourceID, handler);
	}

	/** 
	 *  returns a list of all actions registered for the resourceID.
	 * 
	 * 	<p>
	 *  Guarantees a Action for each element in actionIDs.
	 *  Creates an Action if there is none with the id in the resource map.
	 * 	If there already is an action in the resource map with the id, its 
	 *  eventhandler is replaced by the new one. <p>
	 * 
	 *  PRE: resourceID != null, actionIDs != null, handler != null
	 */

	public List registerHandler(List actionIDs, Object resourceID, Object handler) {
    ensureActionList(resourceID);
		updateActionList(actionIDs, resourceID, handler);
		return getActions(resourceID);
  }

	public Action registerHandler(Object actionID, Object resourceID, Object handler) {
	  List actions = registerHandler(Arrays.asList(new Object[] {actionID}), resourceID, handler);
	  return findAction(actionID, actions);
	}
	
  /** creates and registers actions for all actionIDs to resource and binds
	 *  them to handler.
	 * 	
	 *  @deprecated: old implementation could install duplicates of actionIDs
	 *  - doing it correctly is equivalent to setActionList
	 * 
	 */ 
	public void addToActionList(List actionIDs, Object resourceID, Object handler) {
		ensureActionList(resourceID);
		updateActionList(actionIDs, resourceID, handler);
	}

	/** moves all actions registered in fromResourceID to targetResourceID.
	 *  <p>
	 *  quick hack to enable correct finding of actions in placeholders.
	 * 	PENDING: really support nested actions/containers!
	 * 	POST: list with targetListID added might be empty if sourceListID empty
	 * 	or not found
	 * 	
	 */
	public void moveActions(Object fromResourceID, Object targetResourceID) {
		List targetList = ensureActionList(targetResourceID);
		if (getActions(fromResourceID) == null) return;
		for (Iterator iterator = getActions(fromResourceID).iterator(); iterator.hasNext();) {
			 targetList.add(iterator.next());
		}
		
		removeActionList(fromResourceID);
	}
	

	/** registers and returns the action with actionID in list resourceID.
	 * 	Creates an action if none is found in the list with resourceID. 
	 *  Creates a new list for resourceID if none is available yet. 
	 * 	Registers the action with resource "default" if resourceID == null;
	 * 
	 * <p>
	 * 	PRE: actionID != null.
	 * 
	 */
	public final Action registerAction(Object actionID, Object resourceID) {
		if (resourceID == null) {
			resourceID = "default";
		}
		List resourceList = ensureActionList(resourceID);
		DelegateAction action = findAction(actionID, resourceList);
		if (action == null) {
			action = createDelegateAction(actionID);
			resourceList.add(action);
		}
		return action;
	}

	/** registers an Action as handler. <p>
	 * 
	 *  The returned delegate carries the presentation values,
	 *  delegates the performed(..) to the handler and synchronizes 
	 *  its enabled/disabled state with the handler. 
	 * 
	 * @param actionID  
	 * @param resourceID
	 * @param handler the action to perform
	 * @return the delegate action for use in ui components
	 */
	public Action registerActionAsHandler(Object actionID, Object resourceID, Action handler) {
	  DelegateAction action = (DelegateAction) registerAction(actionID, resourceID);
	  action.addAction(handler);
	  return action;
	}
	/** returns a list of all actions registered for resourceID.
	 * 
	 */
	public List getActions(Object resourceID) {
		return (List) getActionLists().get(resourceID);
	}

	/** returns an action with the id actionID which is registered for
	 * resourceID listID.
	 */
	public Action getAction(Object actionID, Object resourceID) {
		List actions = getActions(resourceID);
		if (actions != null) {
			return findAction(actionID, actions);
		}
		return null;
	}

//---------------------------helper

	/** replaces the handler of all items in actionIDs in the given resourceID.
	 *  <p>
	 * 	creates an action if none with a given actionID is found in resourceID.
	 *  @param actionIDs - a list of actionIDs (_not_ Actions!)
	 *  <p>
	 *  pre: getActionList(resourceID) != null, actionIDs != null.
	 */
	private void updateActionList(List actionIDs, Object resourceID,
			Object handler) {
		List loadedActions = getActions(resourceID);
		for (Iterator iterator = actionIDs.iterator(); iterator.hasNext(); ) {
			Object actionID = iterator.next();
			DelegateAction action = findAction(actionID, loadedActions);
			if (action == null) {
				action = createDelegateAction(actionID);
				loadedActions.add(action);
			}
			replaceHandler(action, handler);
		}
	}

	/** searches and return a DelegateAction with actionID from List of Actions.
	 *  <p> 
	 * 	the comparable is the actionCommand!
	 *  PRE: items in actions have to be of type DelegateActions.
	 *  PRE: actionID != null, actions != null
	 */
	private DelegateAction findAction(Object actionID, List actions) {
		for (Iterator iterator = actions.iterator(); iterator.hasNext(); ) {
			DelegateAction action = (DelegateAction) iterator.next();
			if (actionID.toString().equals(ActionHelper.getActionCommand(action))) {
				return action;
			}
		}
		return null;
	}

	private void replaceHandler(DelegateAction action, Object handler) {
		action.addActionListener(createEventHandlerProxy(handler,
				ActionHelper.getActionCommand(action)), handler);
	}



//---------------------------helper

	protected List ensureActionList(Object resourceID) {
		if (getActions(resourceID) == null) {
			/// arrrghh ... had switched key/value!! before 15.2.2004
			getActionLists().put(resourceID, createList());
		}
		return getActions(resourceID);
	}

	private void removeActionList(Object fromResourceID) {
		getActionLists().remove(fromResourceID);
	}


//---------------------------creating actions and entries
	protected ActionListener createEventHandlerProxy(Object handler,
			String methodName) {
		return (ActionListener) EventHandler.create(ActionListener.class,
				handler, methodName);
	}

	/** creates and returns a DelegateAction with name and actionCommand set
	 *  to actionID's String-rep. 
	 * PRE: actionID != null.
	 * 
	 */
	protected DelegateAction createDelegateAction(Object actionID) {
		DelegateAction action = new DelegateAction(actionID.toString());
		ActionHelper.setActionCommand(action, actionID.toString());
		return action;
	}

//---------------------------init
	// this method needs to be overridden if pluggable subclasses are needed
	private static ActionFactory createActionFactory() {
		return new ActionFactory();
	}

//---------------------------factory methods
	protected Map getActionLists() {
		if (actionLists == null) {
			actionLists = createMap();
		}
		return actionLists;
	}

	protected Map createMap() {
		return new HashMap();
	}

	protected List createList() {
		return new ArrayList();
	}
}
