/*
 * Created on 17.08.2004
 *
 */
package de.kleopatra.view.presentation;

import java.io.*;
import java.net.URL;
import java.text.Collator;
import java.util.*;

import javax.swing.UIManager;

import de.kleopatra.persistence.ResourceManager;
import de.kleopatra.view.UIProperties;


/**
 * Default implementation of localization support. <p>
 * 
 * @author (C) 2004 Jeanette Winzenburg, Berlin
 * @version $Revision: 1.1.1.1 $ - $Date: 2004/08/19 11:18:33 $
 */
public class DefaultBundleHandler implements BundleHandler {

  private Comparator descriptionComparator;

//------------------------ BundleHandler
  
  /** @see BundleHandler. <p>
   * 
   * Here: synch is "all-or-nothing" on a presentation basis - 
   * if the bundle contains a block of entries for the given 
   * presentation the complete block will be overwritten.
   * 
   */
  public void updateBundle(Presentation presentation, URL bundle) {
    if (bundle == null) throw new RuntimeException("URL not available: " + bundle);
    List sortedDescriptions = getSortedDescriptions(presentation);
    List entryLists = getEntryLists(sortedDescriptions);
    try {
      updateEntries(entryLists, presentation.getPresentationID(), bundle);
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  /** @see BundleHandler. <p>
   * 
   *  Here: Synch is "all-or-nothing" on a presentationDescription basis -
   *  if the resourceBundle contains any key for a component, all 
   *  existing attributes are cleared and the keys as contained in the
   *  bundle are set.
   * 
   */
  public void updatePresentation(ResourceBundle bundle, Presentation presentation) {
    if ((presentation.getPresentationID() == null) || (bundle == null))return;
    List presentationEntries = filter(bundle, presentation);
    if (presentationEntries.isEmpty()) return;
    Map descEntries = map(presentationEntries);
    updateDescriptions(descEntries, presentation);
    
  }

  public void localize(Presentation presentation, Class targetClass) {
    String resource = getResourceBundleName(presentation);
    if (resource != null) {
      ResourceBundle bundle = ResourceManager.getInstance()
      	.getResourceBundle(resource, targetClass);
      updatePresentation(bundle, presentation);
    }
    
  }

//-------------------------- helpers - general
  
  /** encapsulates strategy to find an appropriate resourceBundle
   *  name. <p>
   * 
   *  Here: first check the Presentation, if it does not carry a 
   *  resourceBundle, check the UIManager for a shared bundle. The
   *  result may be null. 
   *  Subclasses may override for a different strategy.
   * 
   */
  protected String getResourceBundleName(Presentation presentation) {
    String name = presentation.getResourceBundleName();
    if (name == null) {
      name = (String) UIManager.get(UIProperties.COMMON_RESOURCE_BUNDLE_NAME);
    }
    return name;
  }


  private List getSortedDescriptions(Presentation presentation) {
    List list = createList();
    list.addAll(presentation.getPresentationDescriptions());
    Collections.sort(list, getDescriptionComparator());
    return list;
  }
  
  private List getEntryLists(List sortedDescriptions) {
    List entryLists = createList();
    for (Iterator iter = sortedDescriptions.iterator(); iter.hasNext();) {
      PresentationDescription desc = (PresentationDescription) iter.next();
      List entries = getSortedEntries(desc);
      if (!entries.isEmpty()) {
        entryLists.add(entries);
      }
    }
    return entryLists;
  }

  private List getSortedEntries(PresentationDescription desc) {
    List entries = createList();
    List attributes = desc.getAttributeKeys();
    Collections.sort(attributes);
    for (Iterator iter = attributes.iterator(); iter.hasNext();) {
      String key = (String) iter.next();
      entries.add(createEntry(desc.getPresentationID(), key, desc.getAttribute(key)));
    }
    return entries;
  }

  
  //--------------------------- helpers - form bundle to presentation
  

  private Map map(List presentationEntries) {
	  Map map = createMap();
	  for (Iterator iter = presentationEntries.iterator(); iter.hasNext();) {
      Entry entry = (Entry) iter.next();
      List descEntries = (List) map.get(entry.descName);
      if (descEntries == null) {
        descEntries = createList();
        map.put(entry.descName, descEntries);
      }
      descEntries.add(entry);
    }
    return map;
  }

  private List filter(ResourceBundle bundle, Presentation presentation) {
	  List list = createList();
	  String blockName = presentation.getPresentationID().toString() + ".";
	  for(Enumeration enum = bundle.getKeys(); enum.hasMoreElements();) {
	    String key = (String) enum.nextElement();
	    if (key.startsWith(blockName)) {
	      Entry entry = createEntry(key.substring(blockName.length()), bundle.getString(key));
	      if (entry != null) {
	        list.add(entry);
	      }
	    }
	  }
    return list;
  }

  
  private Entry createEntry(String compoundKey, String value) {
    StringTokenizer st = new StringTokenizer(compoundKey, ".");
    if (st.countTokens() != 2) return null; 
    String descName = st.nextToken();
    String attributeName = st.nextToken();
    return new Entry(descName, attributeName, value);
  }
  
	private void updateDescriptions(Map descEntries, Presentation presentation) {
	  for (Iterator iter = presentation.getPresentationDescriptions().iterator(); iter.hasNext();) {
      PresentationDescription desc = (PresentationDescription) iter.next();
      List entryList = (List) descEntries.get(desc.getPresentationID());
      if (entryList != null) {
        updateDescription(desc, entryList);
      }
      
    }
  }

  private void updateDescription(PresentationDescription desc, List entryList) {
    desc.clearAttributes();
    for (Iterator iter = entryList.iterator(); iter.hasNext();) {
      Entry entry = (Entry) iter.next();
      desc.setAttribute(entry.attributeName, entry.value);
    }
  }

//---------------------------- helpers - from Presentation to bundle (= .properties file)
  
  private void updateEntries(List entryLists, Object formID, URL bundle) throws IOException {
    List unrelatedLines = readUnrelatedLines(formID, bundle);
    writeUnrelatedLinesAndEntries(unrelatedLines, entryLists, formID, bundle);
  }

  private List readUnrelatedLines(Object formID, URL bundle) throws IOException {
    BufferedReader input;
    try {
      input = getInput(bundle);
      
    } catch (FileNotFoundException e) {
      // nothing to read
      return Collections.EMPTY_LIST;
    }
    List lines = createList();
    String startMarker = createMarker(formID, true);
    String endMarker = createMarker(formID, false);
    String line;
    try {
	    while ((line = input.readLine()) != null) {
	      if (line.indexOf(startMarker) < 0) {
	        lines.add(line);
	      } else {
	        skipLinesUntil(endMarker, input);
	      }
	    }
    } finally {
      try {
        input.close();
      } catch (Exception ex){
        ex.printStackTrace();
      }
      
    }
    return lines;
  }


  private void skipLinesUntil(String endMarker, BufferedReader input) throws IOException {
    String line;
    while ((line = input.readLine()) != null) {
      if (line.indexOf(endMarker) >= 0) return;
    }
  }

  public String createMarker(Object formID, boolean startBlock) {
    String marker = "<" + (startBlock ? "" : "/")+ formID + ">";
    return saveConvert(marker, false);
  }

  private void writeUnrelatedLinesAndEntries(List unrelated, List entryLists, 
      Object formID, URL bundle) throws IOException {
    BufferedWriter output = getOutput(bundle);
    try {
      writeEntries(entryLists, formID, output);
      writeLines(unrelated, output);
    } finally {
      try {
        output.close();
      } catch (Exception ex){
        ex.printStackTrace();
      }
    }
  }

  private void writeLines(List unrelated, BufferedWriter output) throws IOException {
    for (Iterator iter = unrelated.iterator(); iter.hasNext();) {
      String line = (String) iter.next();
      writeLine(line, output);
    }
    output.flush();
    
  }

  private void writeEntries(List entryLists, Object formID, BufferedWriter output) throws IOException {
    writeIntro(formID, output);
    for (Iterator iter = entryLists.iterator(); iter.hasNext();) {
      List entries = (List) iter.next();
      writeEntry(entries, formID, output);
    }
    writeFooter(formID, output);
    output.flush();
  }

  private void writeFooter(Object formID, BufferedWriter output) throws IOException {
    output.newLine();
    writeLine("# " + createMarker(formID, false), output);
    
  }

  private void writeIntro(Object formID, BufferedWriter output) throws IOException {
    output.newLine();
    writeLine("# " + createMarker(formID, true), output);
    output.newLine();
    
  }

  private void writeEntry(List entries, Object formID, BufferedWriter output) throws IOException {
    for (Iterator iter = entries.iterator(); iter.hasNext();) {
      Entry entry = (Entry) iter.next();
      writeLine(saveConvert(entry.getBundleRepresentation(formID), false), output);
    }
  }

  private void writeLine(String string, BufferedWriter output) throws IOException {
    output.write(string);
    output.newLine();
    
  }
//  private static final String specialSaveChars = "=: \t\r\n\f#!";
  private static final String specialSaveChars = " \t\r\n\f";
  /** A table of hex digits */
  private static final char[] hexDigit = {
'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'
  };

  /*
   * Converts unicodes to encoded &#92;uxxxx
   * and writes out any of the characters in specialSaveChars
   * with a preceding slash
   */
  private String saveConvert(String theString, boolean escapeSpace) {
      int len = theString.length();
      StringBuffer outBuffer = new StringBuffer(len*2);

      for(int x=0; x<len; x++) {
          char aChar = theString.charAt(x);
          switch(aChar) {
	case ' ':
	    if (x == 0 || escapeSpace) 
		outBuffer.append('\\');

	    outBuffer.append(' ');
	    break;
              case '\\':outBuffer.append('\\'); outBuffer.append('\\');
                        break;
              case '\t':outBuffer.append('\\'); outBuffer.append('t');
                        break;
              case '\n':outBuffer.append('\\'); outBuffer.append('n');
                        break;
              case '\r':outBuffer.append('\\'); outBuffer.append('r');
                        break;
              case '\f':outBuffer.append('\\'); outBuffer.append('f');
                        break;
              default:
                  if ((aChar < 0x0020) || (aChar > 0x007e)) {
                      outBuffer.append('\\');
                      outBuffer.append('u');
                      outBuffer.append(toHex((aChar >> 12) & 0xF));
                      outBuffer.append(toHex((aChar >>  8) & 0xF));
                      outBuffer.append(toHex((aChar >>  4) & 0xF));
                      outBuffer.append(toHex( aChar        & 0xF));
                  } else {
                      if (specialSaveChars.indexOf(aChar) != -1)
                          outBuffer.append('\\');
                      outBuffer.append(aChar);
                  }
          }
      }
      return outBuffer.toString();
  }

  private static char toHex(int nibble) {
  	return hexDigit[(nibble & 0xF)];
      }
  
  private BufferedWriter getOutput(URL bundle) throws IOException {
   Properties props;
    File file = new File(bundle.getFile());
    FileOutputStream outputStream = new FileOutputStream(file);
    OutputStreamWriter fileWriter = new OutputStreamWriter(outputStream, "8859_1");
    return new BufferedWriter(fileWriter);
  }
  
  private BufferedReader getInput(URL bundle) throws IOException {
    File file = new File(bundle.getFile());
    FileInputStream inputStream = new FileInputStream(file);
    InputStreamReader reader = new InputStreamReader(inputStream, "8859_1");
    return new BufferedReader(reader);
  }

//--------------------------- Utility classes
  
  public static class DescriptionComparator implements Comparator {

    public int compare(Object one, Object two) {
      if (!(one instanceof PresentationDescription)
          ||!(two instanceof PresentationDescription)) {
        return 0;
      }
      Object idOne =
        ((PresentationDescription) one).getPresentationID();
      Object idTwo =
        ((PresentationDescription) two).getPresentationID();
      return Collator.getInstance().compare(String.valueOf(idOne),
          String.valueOf(idTwo));
    }

  }
 
  private static class Entry {
    private static String del = ".";
    private static String equal = "=";
    String descName;
    String attributeName;
    String value;
    public Entry(String descName, String attributeName, String value) {
      this.descName = descName;
      this.attributeName = attributeName;
      this.value = value;
    }
    public String getBundleRepresentation(Object formID) {
      return String.valueOf(formID) + del + toString();
    }
    
    public String toString() {
      return descName + del + attributeName + equal + value;
    }
    
  }
 
//------------------------ factory methods + lazy access

  private Comparator getDescriptionComparator() {
    if (descriptionComparator == null) {
      descriptionComparator = new DescriptionComparator();
    }
    return descriptionComparator;
  }

  private Object createEntry(Object descID, String key, String value) {
    return new Entry(String.valueOf(descID), key, value);
  }

  
  private List createList() {
    return new ArrayList();
  }

  private Map createMap() {
    return new HashMap();
  }


}
