
/**
 * @(#)DelegateAction.java      1.6 03/04/16
 *
 * Copyright 2003 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 *
 * - Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * - Redistribution in binary form must reproduce the above
 * copyright notice, this list of conditions and the following
 * disclaimer in the documentation and/or other materials
 * provided with the distribution.
 *
 * Neither the name of Sun Microsystems, Inc. or the names of
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * This software is provided "AS IS," without a warranty of any
 * kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY
 * EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY
 * DAMAGES OR LIABILITIES SUFFERED BY LICENSEE AS A RESULT OF OR
 * RELATING TO USE, MODIFICATION OR DISTRIBUTION OF THIS SOFTWARE OR
 * ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE
 * FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT,
 * SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF
 * THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF SUN HAS
 * BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * You acknowledge that this software is not designed, licensed or
 * intended for use in the design, construction, operation or
 * maintenance of any nuclear facility.
 *
 */

//original: package com.sun.jaf.ui;

package de.kleopatra.view.presentation;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.HashMap;
import java.util.Map;

import javax.swing.AbstractAction;
import javax.swing.Action;

/**
 * Delegates the action performed to another ActionListener.
 *
 * @version  1.6 04/16/03
 * @author  Mark Davidson
 * 
 * Added support for Action as delegate. This will guarantee
 * enabled property to be in synch.
 *  
 * @author Jeanette Winzenburg
 */
public class DelegateAction extends AbstractAction {

	// single instance unicast listener. Can add multicast
	// support by using an EventListenerList.
	private ActionListener listener;
	private Object handler;
	private Action action;
	
	private PropertyChangeListener propertyChangeListener;
	
	public DelegateAction(String name) {
		super(name);
	}

	public DelegateAction() {
		super();
	}

//----------------------------- implementing Action	
	public void actionPerformed(ActionEvent evt) {
		if (listener != null) {
			listener.actionPerformed(evt);
		}
	}

	public void setEnabled(boolean enabled) {
	  setSuperEnabled(enabled);
	  if (action != null) {
	    action.setEnabled(enabled);
	  }
	}
//------------------------------------ accessing
	
	/** add actionListener as delegate and set handler. <p>
	 * 
	 *  PENDING: do we really need access to the handler? 
	 *  internally used for testing only
	 */
	public void addActionListener(ActionListener listener, Object handler) {
		setListenerAndHandler(listener, handler);
	}

	/** add actionListener as delegate and set handler to null.
	 * 
	 * @param listener
	 */
  public void addActionListener(ActionListener listener) {
		setListenerAndHandler(listener, null);
	}
  
  /**
   * add action as actionListener. <p>
   * guarantees synch of enabled property between this and the action.
   * 
   * @param action
   */
  public void addAction(Action action) {
    setListenerAndHandler(null, null);
    if (action == null) return;
    this.action = action;
    listener = action;
    updateEnabledFromDelegate(action.isEnabled());
    action.addPropertyChangeListener(getActionChangeListener());
    
  }
  /** remove actionListener and handler, if any.
   * 	does nothing if the listener is not the same as
   *  the delegate
   */
	public void removeActionListener(ActionListener listener) {
		if ((listener != null) && (listener == this.listener)) {
		  setListenerAndHandler(null, null);
		}
	}

	public ActionListener[] getActionListeners() {
		return new ActionListener[]{ listener };
	}

	/** used in testing only. Should remove?
	 * 
	 * @return
	 */
	public Object getHandler() {
		return handler;
	}
	
	/** convenience: returns list of properties.
	 * 
	 */
	public String toString() {
		return super.toString() + " values: " + String.valueOf(getValueMap());
	}

//---------------------------helper 

	private void setListenerAndHandler(ActionListener listener, Object handler) {
    this.listener = listener;
		this.handler = handler;
		if ((listener == null) && (action != null)) {
		  action.removePropertyChangeListener(propertyChangeListener);
		}
  }

// ----------------  synch support
	
  private PropertyChangeListener getActionChangeListener() {
    if (propertyChangeListener == null) {
      propertyChangeListener = createActionChangeListener();
    }
    return propertyChangeListener;
  }

  private PropertyChangeListener createActionChangeListener() {
    PropertyChangeListener l = new PropertyChangeListener() {

      public void propertyChange(PropertyChangeEvent evt) {
        if ("enabled".equals(evt.getPropertyName())) {
          updateEnabledFromDelegate(((Boolean) evt.getNewValue()).booleanValue());
        }
        
      }
      
    };
    return l;
  }

  protected void updateEnabledFromDelegate(boolean delegateEnabled) {
    if (delegateEnabled == isEnabled()) return;
    setSuperEnabled(delegateEnabled);
  }

  private void setSuperEnabled(boolean delegateEnabled) {
    super.setEnabled(delegateEnabled);
    
  }

	
	protected Map getValueMap() {
		Map values = new HashMap();
		Object[] keys = getKeys();
		for (int i = 0; i < keys.length; i++) {
			values.put(keys[i], getValue(keys[i].toString()));
		}
		values.put("listener", (listener != null) ? listener.getClass() : null);
		values.put("handler", handler);
		return values;
	}
}
