package de.kleopatra.view.table;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;

/**	Make Table behave.
 *
 *	<p> handles
 *	<ul>
 *  <li> column changes (updates only)
 *  <li> selection changes via mouseEvents in header
 *  <li> does not use keystrokes with modifiers (except shift) to start edit
 *  <li> registers correct action to cancel edits (via editor.cancelEditing() 
 * 		instead of table.removeEditor()
 *  <li> terminates edit on column change notification (brute force - should do
 * 	if needed only)
 *  <li> Removed again ... usability issues in FormBuilder:   
 *  [couple anchor mark to JTable child of active window 
 * 	instead of to JTable focus]
 *	</ul>
 *
 * @author (C) Jeanette Winzenburg, Berlin
 * @version $Revision: 1.2 $ - $Date: 2004/09/30 10:18:01 $
 *
 */
public class JExTable extends JTable {

	public static final String PREFERRED_VISIBLE_COLUMNS_KEY =
		"JExtTable.preferredVisibleColumns";
//	private static final int ROWS = 10;
//	private static final int COLUMNS = 3;
	// supports table selection on mouseEvents of header
	// pending: interference with resizing/moving columns?
	protected HeaderSelection headerSelection;
	private boolean ensureFillViewport = true;

	public JExTable() {}
	
	public JExTable(TableModel model) {
	  super(model);
	}

//	public JExTable(int rows, int columns) {
//	  super(rows, columns);
//	}
	//---------------------------overriding super

	public Dimension getPreferredScrollableViewportSize() {
		Integer visibleColumns =
			(Integer) getClientProperty(PREFERRED_VISIBLE_COLUMNS_KEY);
		Dimension pref = getPreferredSize();
		if (visibleColumns != null) {
			int width = 0;
			for (int i = 0;
				i < Math.min(visibleColumns.intValue(), getColumnCount());
				i++) {
				width += getColumnModel().getColumn(i).getPreferredWidth();
			}
			if ((width > 0) && (width < pref.width)) {
				pref.width = width;
			}
		}
		return pref;
	}

	public void setModel(TableModel model) {
		super.setModel(model);
	}

	public void setTableHeader(JTableHeader header) {
		super.setTableHeader(header);
		getHeaderSelection().setTableHeader(header);
	}

	public void tableChanged(TableModelEvent e) {
		if (isColumnUpdated(e)) {
			updateColumnHeader(e.getColumn());
			// reacting to insert/remove is buggy - abort for now!
			/*} else if (isColumnInserted(e)) {
				insertColumnHeader(e.getColumn());
			} else if (isColumnRemoved(e)) {
				removeColumnHeader(e.getColumn());
			*/
		} else {
		  ensureValidEditing(e);
			super.tableChanged(e);
			if (mightBeColumn(e)) {
				updateColumnHeader(e.getColumn());
			}
			if (isStructureChanged(e) && getAutoCreateColumnsFromModel()) {
			  adjustColumnWidths();
			}
		}
	}

	
  public Component prepareRenderer(TableCellRenderer renderer, int row,
      int column) {
    Object value = getValueAt(row, column);
  	boolean isSelected = isCellSelected(row, column);
  	boolean hasFocus = shouldRenderCellFocused(row, column);

  	return renderer.getTableCellRendererComponent(this, value,
  	                                              isSelected, hasFocus,
  	                                              row, column);
    
  }

  protected boolean shouldRenderCellFocused(int row, int column) {
    boolean rowIsAnchor = (selectionModel.getAnchorSelectionIndex() == row);
  	boolean colIsAnchor =
  	    (columnModel.getSelectionModel().getAnchorSelectionIndex() == column);
  	boolean hasFocus = (rowIsAnchor && colIsAnchor) && shouldRenderCellFocused();
    return hasFocus;
  }
 
  protected boolean shouldRenderCellFocused() {
    // is super behaviour... THINK
    return hasFocus();
//    // PENDING: costly to do on every call to prepareRenderer?
//    // could store an listen to container events?
//    Window window = SwingUtilities.windowForComponent(this);
//      //KeyboardFocusManager.getCurrentKeyboardFocusManager().getActiveWindow();
//    return window != null ? window.isActive() : false;

  }
  
  //---------------------- take control about columnModelEvents from super
// PENDING: all except columnSelectionChanged.
	


  private boolean isStructureChanged(TableModelEvent e) {
    return e == null || e.getFirstRow() == TableModelEvent.HEADER_ROW;
  }

  /** hook for subclasses to do something about tableWidths after
   *  structure changes.
   *
   */
  protected void adjustColumnWidths() {
    
  }

  public void columnMarginChanged(ChangeEvent e) {
    terminateEditing();
  	TableColumn resizingColumn = getResizingColumn();
  	// Need to do this here, before the parent's
  	// layout manager calls getPreferredSize().
  	if (resizingColumn != null && autoResizeMode == AUTO_RESIZE_OFF) {
  	    resizingColumn.setPreferredWidth(resizingColumn.getWidth());
  	}
  	resizeAndRepaint();
  }
  

	public void columnAdded(TableColumnModelEvent e) {
    terminateEditing();
    resizeAndRepaint();
  }

  public void columnMoved(TableColumnModelEvent e) {
    terminateEditing();
    repaint();
  }
  
  public void columnRemoved(TableColumnModelEvent e) {
    if (e.getFromIndex() == getEditingColumn()) {
      cancelEditing();
    } else {
      terminateEditing();
    }
    resizeAndRepaint();
  }

//----------------------helper for fine-grained control about terminating edits
  

  /** brute force bug fix: table does nothing to
	 *  terminate edits in tableChanged.
	 * @param e
	 */
	protected void ensureValidEditing(TableModelEvent e) {
	  cancelEditing();
  }

  protected void cancelEditing() {
    if (!isEditing()) return;
    getCellEditor().cancelCellEditing();
    
  }


  protected void terminateEditing() {
    if (!isEditing()) return;
    if (!getCellEditor().stopCellEditing()) {
      getCellEditor().cancelCellEditing();
    }
    
  }

  private TableColumn getResizingColumn() {
  	return (tableHeader == null) ? null
  	                             : tableHeader.getResizingColumn();
      }

  
  private boolean processKeyBindingCopied(
		KeyStroke ks,
		KeyEvent e,
		int condition,
		boolean pressed) {
		InputMap map = getInputMap(condition);
		ActionMap am = getActionMap();
		if ((map != null) && (am != null) && isEnabled()) {
			Object binding = map.get(ks);
			Action action = (binding == null) ? null : am.get(binding);
			if (action != null) {
				return SwingUtilities.notifyAction(
					action,
					ks,
					e,
					this,
					e.getModifiers());
			}
		}
		return false;
	}

	protected boolean processKeyBinding(
		KeyStroke ks,
		KeyEvent e,
		int condition,
		boolean pressed) {
		// quick-hack to prevent (most) accelerators from starting edit
		// PENDING... really solve!
		if (shouldTryStartingEdit(e)) {
			return super.processKeyBinding(ks, e, condition, pressed);
		}
		return processKeyBindingCopied(ks, e, condition, pressed);
	}

	private boolean shouldTryStartingEdit(KeyEvent e) {
		boolean shouldTry = (e.getModifiers() == 0) || (e.getModifiers() == KeyEvent.SHIFT_DOWN_MASK)
			|| (e.getModifiers() == KeyEvent.SHIFT_MASK);
		return shouldTry && !e.isActionKey();
	}

	// this is copied and changed as indicated
/*	public void changeSelection(int rowIndex, int columnIndex, boolean toggle, boolean extend) {
		// Autoscrolling support.
		if (getAutoscrolls()) {
			Rectangle cellRect = getCellRect(rowIndex, columnIndex, false);
			if (cellRect != null) {
				scrollRectToVisible(cellRect);
			}
		}

		ListSelectionModel rsm = getSelectionModel();
		ListSelectionModel csm = getColumnModel().getSelectionModel();
		changeSelectionModel(rsm, rowIndex, toggle, extend);
		//CHANGED:
		if (cellSelectionEnabled) {
			// this automatically selects the colum
			// you'll need some gesture external to the JTable to
			// deselect columns (f.i. in JTableHeader)
			csm.addSelectionInterval(columnIndex, columnIndex);
		} else {
			changeSelectionModel(csm, columnIndex, toggle, extend);
		}


	}
	//this is copied (because its private)
	private void changeSelectionModel(ListSelectionModel sm, int index,
																		boolean toggle, boolean extend) {
		if (extend) {
			if (toggle) {
				sm.setAnchorSelectionIndex(index);
			}
			else {
				sm.setLeadSelectionIndex(index);
			}
		}
		else {
			if (toggle) {
				if (sm.isSelectedIndex(index)) {
					sm.removeSelectionInterval(index, index);
				}
				else {
					sm.addSelectionInterval(index, index);
				}
			}
			else {
				sm.setSelectionInterval(index, index);
			}
		}
	}
*/
	//---------------------------init

	protected void initializeLocalVars() {
		super.initializeLocalVars();
		setSurrendersFocusOnKeystroke(true);
		putClientProperty("terminateEditOnFocusLost", Boolean.TRUE);
		// autoStartsEdit is the default, set it anyway...
		putClientProperty("JTable.autoStartsEdit", Boolean.TRUE);
		setAutoResizeMode(AUTO_RESIZE_OFF);
		setDefaultEditor(Object.class, createDefaultObjectEditor());
		adjustActionMap();
	}

	//---------------------------helper

    
  

  /** Workaround for buggy default CancelAction (1.4/1.5).
	 *  default will call table.removeEditor instead of cancelling edits
	 *  via editro.cancelCellEditing().
	 */
	protected void adjustActionMap() {
		AbstractAction cancel = new AbstractAction() {

			public void actionPerformed(ActionEvent e) {
				
				if (isEditing()) {
					getCellEditor().cancelCellEditing();
				}
				
			}
			
		};
		getActionMap().put("cancel", cancel);
		
	}

	protected void updateColumnHeader(int modelColumn) {
		int viewColumn = convertColumnIndexToView(modelColumn);
		if ((viewColumn < 0) || (viewColumn >= getColumnModel().getColumnCount())) return; 
		TableColumn column = getColumnModel().getColumn(viewColumn);
	//	column.setHeaderValue(getModel().getColumnName(column.getModelIndex()));
		setHeaderValue(column);
		getTableHeader().repaint();
	}

	protected void setHeaderValue(TableColumn column) {
		column.setHeaderValue(getModel().getColumnName(column.getModelIndex()));
	}

	/*
	 	// something wrong here...
	 	protected void insertColumnHeader(int modelColumn) {
			int viewColumn = modelColumn; //convertColumnIndexToView(modelColumn);
			TableColumn column = createTableColumn(modelColumn);
			column.setHeaderValue(getModel().getColumnName(modelColumn));
			getColumnModel().addColumn(column);
			getColumnModel().moveColumn(getColumnCount() - 1, viewColumn);
		}
	
		// not yet testet!!
		protected void removeColumnHeader(int modelColumn) {
			int viewColumn = convertColumnIndexToView(modelColumn);
			TableColumn column = getColumnModel().getColumn(viewColumn);
			getColumnModel().removeColumn(column);
		}
	
	*/
	protected boolean isColumnUpdated(TableModelEvent e) {
		return isColumn(e, TableModelEvent.UPDATE);
	}

	protected boolean isColumnInserted(TableModelEvent e) {
		return isColumn(e, TableModelEvent.INSERT);
	}

	protected boolean isColumnRemoved(TableModelEvent e) {
		return isColumn(e, TableModelEvent.DELETE);
	}

	protected boolean isColumn(TableModelEvent e, int type) {
		return isColumn(e) && (e.getType() == type);
	}

	protected boolean isColumn(TableModelEvent e) {
		int firstRow = e.getFirstRow();
		int lastRow = e.getLastRow();
		return (e.getColumn() != TableModelEvent.ALL_COLUMNS)
			&& (firstRow != lastRow)
			// PENDING: problem with rowCount == 1, can't tell then
			&& (firstRow == 0)
			&& (lastRow == getRowCount() - 1);
	}

	protected boolean mightBeColumn(TableModelEvent e) {
		int firstRow = e.getFirstRow();
		int lastRow = e.getLastRow();
		return (e.getColumn() != TableModelEvent.ALL_COLUMNS)
		//	&& (firstRow != lastRow)
			// PENDING: problem with rowCount == 1, can't tell then
			&& (firstRow == 0)
			&& (lastRow == getRowCount() - 1);
		
	}
	//---------------------------factory methods

	private TableCellEditor createDefaultObjectEditor() {
		return new FixedGenericCellEditor();
	}

	protected HeaderSelection getHeaderSelection() {
		if (headerSelection == null) {
			headerSelection = createHeaderSelection();
		}
		return headerSelection;
	}

	protected HeaderSelection createHeaderSelection() {
		return new HeaderSelection();
	}

	protected TableColumn createTableColumn(int modelIndex) {
		return new TableColumn(modelIndex);
	}

	protected JTableHeader createDefaultTableHeader() {
		JTableHeader header = new JTableHeader(columnModel) {
			public void columnSelectionChanged(ListSelectionEvent e) {
				repaint();
			}
			public Dimension getPreferredSize() {
				Dimension pref = super.getPreferredSize();
				Object headerHeight = getClientProperty("JTableHeader.preferredHeight");
				if (headerHeight instanceof Integer) {
					pref.height = ((Integer) headerHeight).intValue();
				}
				return pref;
			}
		};
		return header;
	}

}
